"""
Implementation of an actual reaction network interface.
"""

from typing import List, Optional

from graph_tool.util import find_edge, find_vertex

from rxn_network.core import CostFunction, Enumerator, Network
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.network.gt import (
    initialize_graph,
    update_vertex_props,
    yens_ksp,
)
from rxn_network.network.entry import NetworkEntry, NetworkEntryType
from rxn_network.network.utils import get_loopback_edges, get_rxn_nodes_and_edges
from rxn_network.pathways.basic import BasicPathway
from rxn_network.reactions.reaction_set import ReactionSet


class ReactionNetwork(Network):
    """
    Main reaction network class for building graphs of reactions and performing
    pathfinding.
    """

    def __init__(
        self,
        entries: GibbsEntrySet,
        enumerators: List[Enumerator],
        cost_function: CostFunction,
        open_elem: Optional[str] = None,
        chempot: Optional[float] = None,
    ):
        """
        Initialize a ReactionNetwork object for a set of entires, enumerator,
        and cost function. The network can be constructed by calling build().

        Args:
            entries: iterable of entry-like objects
            enumerators: iterable of enumerators which will be called during the
                build of the network
            cost_function: the function used to calculate the cost of each reaction edge
            open_elem: Optional name of an element that is kept open during reaction
            chempot: Optional associated chemical potential of open element
        """
        super().__init__(
            entries=entries, enumerators=enumerators, cost_function=cost_function
        )
        self.open_elem = open_elem
        self.chempot = chempot

    def build(self):
        """
        Construct the reaction network graph object and store under the "graph"
        attribute. Does NOT initialize precursors or target; you must call set_precursors()
        or set_target() to do so.

        Returns: None

        """
        rxn_set = self._get_rxns()
        costs = rxn_set.calculate_costs(self.cost_function)
        rxns = rxn_set.get_rxns(self.open_elem, self.chempot)

        self.logger.info("Building graph from reactions...")
        nodes, rxn_edges = get_rxn_nodes_and_edges(rxns)

        g = initialize_graph()
        g.add_vertex(len(nodes))
        for i, network_entry in enumerate(nodes):
            props = {"entry": network_entry, "type": network_entry.description.value}
            update_vertex_props(g, g.vertex(i), props)

        edge_list = []
        for edge, cost, rxn in zip(rxn_edges, costs, rxns):
            v1 = g.vertex(edge[0])
            v2 = g.vertex(edge[1])
            edge_list.append((v1, v2, cost, rxn, "reaction"))

        edge_list.extend(get_loopback_edges(g, nodes))

        g.add_edge_list(edge_list, eprops=[g.ep["cost"], g.ep["rxn"], g.ep["type"]])

        self._g = g

    def find_pathways(self, targets: List[str], k: float = 15) -> List[BasicPathway]:
        """

        Find the k-shortest paths to a provided list of 1 or more targets.

        Args:
            targets: List of the formulas of each target
            k: Number of shortest paths to find for each target

        Returns:
            List of BasicPathway objects to all provided targets.
        """
        if not self.precursors:
            raise AttributeError("Must call set_precursors() before pathfinding!")

        paths = []
        for target in targets:
            if type(target) == str:
                target = self.entries.get_min_entry_by_formula(target)
            self.set_target(target)
            print(f"PATHS to {target.composition.reduced_formula} \n")
            print("--------------------------------------- \n")
            pathways = self._shortest_paths(k=k)
            paths.extend(pathways)

        return paths

    def set_precursors(self, precursors):
        """

        Args:
            precursors:

        Returns:

        """
        g = self._g

        precursors = set(precursors)
        if precursors == self.precursors:
            return
        elif self.precursors:
            precursors_v = find_vertex(
                g, g.vp["type"], NetworkEntryType.Precursors.value
            )[0]
            g.remove_vertex(precursors_v)
            loopback_edges = find_edge(g, g.ep["type"], "loopback_precursors")
            for e in loopback_edges:
                g.remove_edge(e)
        elif not all([p in self.entries for p in precursors]):
            raise ValueError("One or more precursors are not included in network!")

        precursors_v = g.add_vertex()
        precursors_entry = NetworkEntry(precursors, NetworkEntryType.Precursors)
        props = {"entry": precursors_entry, "type": precursors_entry.description.value}
        update_vertex_props(g, precursors_v, props)

        add_edges = []
        for v in g.vertices():
            entry = g.vp["entry"][v]
            if not entry:
                continue
            if entry.description.value == NetworkEntryType.Reactants.value:
                if entry.entries.issubset(precursors):
                    add_edges.append((precursors_v, v, 0.0, None, "precursors"))
            elif entry.description.value == NetworkEntryType.Products.value:
                for v2 in g.vertices():
                    entry2 = g.vp["entry"][v2]
                    if entry2.description.value == NetworkEntryType.Reactants.value:
                        if precursors.issuperset(entry2.entries):
                            continue
                        if precursors.union(entry.entries).issuperset(entry2.entries):
                            add_edges.append((v, v2, 0.0, None, "loopback_precursors"))

        g.add_edge_list(add_edges, eprops=[g.ep["cost"], g.ep["rxn"], g.ep["type"]])

        self.precursors = precursors

    def set_target(self, target):
        """

        Args:
            target:

        Returns:

        """
        g = self._g
        if target == self.target:
            return
        elif self.target or target is None:
            target_v = find_vertex(g, g.vp["type"], NetworkEntryType.Target.value)[0]
            g.remove_vertex(target_v)

        target_v = g.add_vertex()
        target_entry = NetworkEntry([target], NetworkEntryType.Target)
        props = {"entry": target_entry, "type": target_entry.description.value}
        update_vertex_props(g, target_v, props)

        add_edges = []
        for v in g.vertices():
            entry = g.vp["entry"][v]
            if not entry:
                continue
            if entry.description.value != NetworkEntryType.Products.value:
                continue
            if target in entry.entries:
                add_edges.append([v, target_v, 0.0, None, "target"])

        g.add_edge_list(add_edges, eprops=[g.ep["cost"], g.ep["rxn"], g.ep["type"]])

        self.target = target

    def _shortest_paths(self, k=15):
        """Finds the k shortest paths using Yen's algorithm and returns BasicPathways"""
        g = self._g
        paths = []

        precursors_v = find_vertex(g, g.vp["type"], NetworkEntryType.Precursors.value)[
            0
        ]
        target_v = find_vertex(g, g.vp["type"], NetworkEntryType.Target.value)[0]

        for path in yens_ksp(g, k, precursors_v, target_v):
            paths.append(self._path_from_graph(g, path))

        for path in paths:
            print(path, "\n")

        return paths

    def _get_rxns(self) -> ReactionSet:
        "Gets reaction set by running all enumerators"
        rxns = []
        for enumerator in self.enumerators:
            rxns.extend(enumerator.enumerate(self.entries))

        rxns = ReactionSet.from_rxns(rxns, self.entries)
        return rxns

    @staticmethod
    def _path_from_graph(g, path):
        rxns = []
        costs = []

        for step, v in enumerate(path):
            if g.vp["type"][v] == NetworkEntryType.Products.value:
                e = g.edge(path[step - 1], v)

                rxns.append(g.ep["rxn"][e])
                costs.append(g.ep["cost"][e])

        return BasicPathway(reactions=rxns, costs=costs)

    @property
    def graph(self):
        return self._g

    @property
    def chemsys(self):
        return "-".join(sorted(self.entries.chemsys))

    def __repr__(self):
        return (
            f"ReactionNetwork for chemical system: "
            f"{self.chemsys}, "
            f"with Graph: {str(self._g)}"
        )



"""
Experimental Gibbs free energy data from NIST-JANAF (compounds, gases) and FactSage (elements)
"""
from pathlib import Path

from monty.serialization import loadfn

cwd = Path(__file__).parent.resolve()

G_COMPOUNDS = loadfn(cwd / "compounds.json")
G_ELEMS = loadfn(cwd / "elements.json")
G_GASES = loadfn(cwd / "gases.json")



"""
This module implements two types of basic reaction enumerators, with or without the
option of an open entry
"""
from itertools import combinations, product
from math import comb
from typing import List, Optional

from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.analysis.phase_diagram import PhaseDiagram, GrandPotentialPhaseDiagram
from tqdm.auto import tqdm

from rxn_network.core import Enumerator
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.enumerators.utils import (
    apply_calculators,
    filter_entries_by_chemsys,
    get_elems_set,
    group_by_chemsys,
    initialize_calculators,
    initialize_entry,
)
from rxn_network.reactions import ComputedReaction
from rxn_network.utils import limited_powerset


class BasicEnumerator(Enumerator):
    """
    Enumerator for finding all simple reactions within a set of entries, up to a
    maximum reactant and product cardinality (n).
    """

    def __init__(
        self,
        precursors: Optional[List[str]] = None,
        target: Optional[str] = None,
        calculators: Optional[List[str]] = None,
        n: int = 2,
        remove_unbalanced: bool = True,
        remove_changed: bool = True,
    ):
        """
        Supplied target and calculator parameters are automatically initialized as
        objects during enumeration.

        Args:
            precursors: Optional list of precursor formulas; only reactions
                which contain at least these phases as reactants will be enumerated.
            target: Optional formula of target; only reactions which include
                formation of this target will be enumerated.
            calculators: Optional list of Calculator object names; see calculators
                module for options (e.g., ["ChempotDistanceCalculator])
            n: Maximum reactant/product cardinality; i.e., largest possible number of
                entries on either side of the reaction. Defaults to 2.
            remove_unbalanced: Whether to remove reactions which are unbalanced.
                Defaults to True.
            remove_changed: Whether to remove reactions which can only be balanced by
                removing a reactant/product or having it change sides. Defaults to True.
        """
        super().__init__(precursors, target, calculators)

        self.n = n
        self.remove_unbalanced = remove_unbalanced
        self.remove_changed = remove_changed

        self._stabilize = False
        if "ChempotDistanceCalculator" in self.calculators:
            self._stabilize = True

        self._build_pd = False
        self._build_grand_pd = False

    def enumerate(self, entries: GibbsEntrySet) -> List[ComputedReaction]:
        """
        Calculate all possible reactions given a set of entries. If the enumerator
        was initialized with specified precursors or target, the reactions will be
        filtered by these constraints. Every enumerator follows a standard procedure:

        1. Initialize entries, i.e. ensure that precursors and target are considered
        stable entries within the entry set. If using ChempotDistanceCalculator,
        ensure that entries are filtered by stability.

        2. Get a dictionary representing every possible node, i.e. phase combination,
        grouped by chemical system.

        3. Filter the combos dictionary for chemical systems which are not relevant (
        i.e. don't contain elements in precursors and/or target.

        4. Iterate through each chemical system, initializing calculators,
        and computing all possible reactions for reactant/product pair and/or
        thermodynamically predicted reactions for given reactants.

        5. Add reactions to growing list, repeat Step 4 until combos dict exhausted.

        Args:
            entries: the set of all entries to enumerate from
        """
        entries, precursors, target = self._get_initialized_entries(entries)
        combos_dict = self._get_combos_dict(entries, precursors, target)
        open_combos = self._get_open_combos(entries)

        pbar = tqdm(combos_dict.items(), desc=self.__class__.__name__)

        rxns = []
        for chemsys, combos in pbar:
            pbar.set_description(f"{chemsys}")

            filtered_entries = filter_entries_by_chemsys(entries, chemsys)
            calculators = initialize_calculators(self.calculators, filtered_entries)

            rxn_iter = self._get_rxn_iterable(combos, open_combos)
            r = self._get_rxns(
                rxn_iter, precursors, target, calculators, filtered_entries
            )
            rxns.extend(r)

        return list(set(rxns))

    def estimate_max_num_reactions(self, entries: List[ComputedEntry]) -> int:
        """
        Estimate the upper bound of the number of possible reactions. This will
        correlate with the amount of time it takes to enumerate reactions.

        Args:
            entries: A list of all entries to consider

        Returns: The upper bound on the number of possible reactions
        """
        return sum([comb(len(entries), i) for i in range(1, self.n + 1)]) ** 2

    def _get_combos_dict(self, entries, precursor_entries, target_entries):
        precursor_elems = (
            get_elems_set(precursor_entries) if precursor_entries else set()
        )
        target_elems = (
            {e.name for e in target_entries.composition.elements}
            if target_entries
            else set()
        )

        combos = [set(c) for c in limited_powerset(entries, self.n)]

        combos_dict = group_by_chemsys(combos)
        filtered_combos = self._filter_dict_by_elems(
            combos_dict, precursor_elems, target_elems
        )

        return filtered_combos

    def _get_open_combos(self, entries):
        """ No open entries for BasicEnumerator, returns None"""
        return None

    def _get_rxns(
        self, rxn_iterable, precursors, target, calculators, filtered_entries=None
    ):
        """
        Returns reactions from an iterable representing the combination of 2 sets
        of phases. Works for reactions with open phases.
        """
        pd = None
        if self.build_pd:
            pd = PhaseDiagram(filtered_entries)

        grand_pd = None
        if self.build_grand_pd:
            chempots = getattr(self, "chempots")
            grand_pd = GrandPotentialPhaseDiagram(filtered_entries, chempots)

        rxns = []
        for reactants, products in rxn_iterable:
            r = set(reactants) if reactants else set()
            p = set(products) if products else set()
            all_phases = r | p

            if (r & p) or (precursors and not precursors.issubset(all_phases)):
                continue

            suggested_rxns = self._react(r, p, calculators, pd, grand_pd)

            for rxn in suggested_rxns:
                if (
                    rxn.is_identity
                    or (self.remove_unbalanced and not rxn.balanced)
                    or (self.remove_changed and rxn.lowest_num_errors != 0)
                ):
                    continue

                if not target or target in rxn.product_entries:
                    if not precursors or precursors.issubset(rxn.reactant_entries):
                        rxns.append(rxn)

        return rxns

    def _react(self, reactants, products, calculators, pd=None, grand_pd=None):
        forward_rxn = ComputedReaction.balance(reactants, products)
        backward_rxn = forward_rxn.reverse()

        forward_rxn = apply_calculators(forward_rxn, calculators)
        backward_rxn = apply_calculators(backward_rxn, calculators)
        return [forward_rxn, backward_rxn]

    def _get_rxn_iterable(self, combos, open_combos):
        return combinations(combos, 2)

    def _get_initialized_entries(self, entries):
        precursors, target = None, None
        entries_new = entries.copy()

        if self.precursors:
            precursors = {
                initialize_entry(f, entries, self.stabilize) for f in self.precursors
            }
            for p in precursors:
                if p not in entries:
                    old_entry = entries_new.get_min_entry_by_formula(
                        p.composition.reduced_formula
                    )
                    entries_new.remove(old_entry)
                    entries_new.add(p)

        if self.target:
            target = initialize_entry(self.target, entries, self.stabilize)
            if self.stabilize:
                entries_new.remove(entries.get_min_entry_by_formula(self.target))
                entries_new.add(target)

        if self.stabilize:
            entries_new = entries_new.filter_by_stability(e_above_hull=0.0)
            self.logger.info("Filtering by stable entries!")

        return entries_new, precursors, target

    @staticmethod
    def _filter_dict_by_elems(combos_dict, precursor_elems, target_elems):
        filtered_dict = dict()
        for chemsys, combos in combos_dict.items():
            elems = chemsys.split("-")
            if (
                (target_elems and not target_elems.issubset(elems))
                or (precursor_elems and not precursor_elems.issubset(elems))
                or len(elems) >= 10
                or len(elems) == 1
            ):
                continue
            else:
                filtered_dict[chemsys] = combos

        return filtered_dict

    @property
    def stabilize(self):
        """ Whether or not to use only stable entries in analysis"""
        return self._stabilize

    @property
    def build_pd(self):
        """Whether or not to build a PhaseDiagram object during reaction enumeration (
        useful for some analyses)"""
        return self._build_pd

    @property
    def build_grand_pd(self):
        """Whether or not to build a GrandPotentialPhaseDiagram object during
        reaction enumeration (useful for some analyses)"""
        return self._build_grand_pd


class BasicOpenEnumerator(BasicEnumerator):
    """
    Enumerator for finding all simple reactions within a set of entries, up to a
    maximum reactant/product cardinality (n), with any number of open phases. Note:
    this does not return OpenComputedReaction objects (this can be calculated using
    the ReactionSet class).
    """

    def __init__(
        self,
        open_phases: List[str],
        precursors: Optional[List[str]] = None,
        target: Optional[str] = None,
        calculators: Optional[List[str]] = None,
        n: int = 2,
        remove_unbalanced: bool = True,
        remove_changed: bool = True,
    ):
        """
        Supplied target and calculator parameters are automatically initialized as
        objects during enumeration.

        Args:
            open_phases: List of formulas of open entries (e.g. ["O2"])
            precursors: Optional list of formulas of precursor phases; only reactions
                which have these phases as reactants will be enumerated.
            target: Optional formula of target; only reactions which make this target
                will be enumerated.
            calculators: Optional list of Calculator object names; see calculators
                module for options (e.g., ["ChempotDistanceCalculator]).
            n: Maximum reactant/product cardinality; i.e., largest possible number of
                entries on either side of the reaction.
            remove_unbalanced: Whether to remove reactions which are unbalanced.
                Defaults to True
            remove_changed: Whether to remove reactions which can only be balanced by
                removing a reactant/product or having it change sides. Defaults to True.
        """
        super().__init__(
            precursors, target, calculators, n, remove_unbalanced, remove_changed
        )
        self.open_phases = open_phases

    def estimate_max_num_reactions(self, entries: List[ComputedEntry]) -> int:
        """
        Estimate the upper bound of the number of possible reactions. This will
        correlate with the amount of time it takes to enumerate reactions.

        Args:
            entries: A list of all entries to consider

        Returns: The upper bound on the number of possible reactions
        """
        num_open_phases = len(self.open_phases)
        num_combos_with_open = sum(
            [comb(num_open_phases, i) for i in range(1, num_open_phases + 1)]
        )

        num_total_combos = 0
        for i in range(1, self.n + 1):
            num_combos = comb(len(entries), i)
            num_total_combos += num_combos_with_open * num_combos

        return num_total_combos ** 2

    def _get_open_combos(self, entries):
        open_entries = {
            e for e in entries if e.composition.reduced_formula in self.open_phases
        }
        open_combos = [
            set(c) for c in limited_powerset(open_entries, len(open_entries))
        ]
        return open_combos

    def _get_rxn_iterable(self, combos, open_combos):
        combos_with_open = [
            combo | open_combo
            for combo in combos
            for open_combo in open_combos
            if not combo & open_combo
        ]
        rxn_iter = product(combos, combos_with_open)

        return rxn_iter



"""
Basic interface for a (reaction) Network
"""
import logging
from abc import ABCMeta, abstractmethod
from typing import List

from monty.json import MSONable
from pymatgen.entries import Entry

from rxn_network.core.pathway import Pathway


class Network(MSONable, metaclass=ABCMeta):
    """
    Base definition for a reaction network
    """

    def __init__(self, entries: List[Entry], enumerators, cost_function):
        self.logger = logging.getLogger(str(self.__class__.__name__))
        self.logger.setLevel("INFO")
        self.entries = entries
        self.enumerators = enumerators
        self.cost_function = cost_function
        self._g = None
        self.precursors = None
        self.target = None

    @abstractmethod
    def build(self):
        """Construct the network from the supplied enumerators"""

    @abstractmethod
    def find_pathways(self, target, k) -> List[Pathway]:
        """Find reaction pathways"""

    @abstractmethod
    def set_precursors(self):
        """Set the phases used as precursors in the network"""

    @abstractmethod
    def set_target(self):
        """Set the phase used as a target in the network"""



""" Tests for NistReferenceEntry. """
import pytest

from pymatgen.core.composition import Composition

from rxn_network.entries.nist import NISTReferenceEntry


@pytest.fixture(scope="session")
def entries():
    comp = Composition("CO2")
    temps = [300, 600, 900, 1200, 1500, 1800]
    return {t: NISTReferenceEntry(composition=comp, temperature=t) for t in temps}


def test_invalid_formula():
    with pytest.raises(ValueError) as error:
        assert NISTReferenceEntry(Composition("AX"), temperature=300)
    assert str(error.value) == "Formula must be in NIST-JANAF thermochemical tables"


def test_invalid_temperature():
    with pytest.raises(ValueError) as error:
        assert NISTReferenceEntry(Composition("K2CO3"), temperature=200)
    assert str(error.value) == "Temperature must be selected from range: [300, 2000] K"

    with pytest.raises(ValueError) as error:
        assert NISTReferenceEntry(Composition("K2CO3"), temperature=2300)
    assert str(error.value) == "Temperature must be selected from range: [300, 2000] K"


def test_energy(entries):
    expected_energies = [
        -4.087606831162386,
        -4.095773877778095,
        -4.101640055931004,
        -4.105267551255239,
        -4.107236763002682,
        -4.107910440705754,
    ]
    actual_energies = [entry.energy for entry in entries.values()]

    assert actual_energies == pytest.approx(expected_energies)


def test_energy_per_atom(entries):
    expected_energies = [
        -1.362535610387462,
        -1.365257959259365,
        -1.3672133519770011,
        -1.3684225170850797,
        -1.369078921000894,
        -1.3693034802352513,
    ]
    actual_energies = [entry.energy_per_atom for entry in entries.values()]

    assert actual_energies == pytest.approx(expected_energies)


def test_correction_uncertainty(entries):
    assert all([e.correction_uncertainty == 0 for e in entries.values()])


def test_correction_uncertainty_per_atom(entries):
    assert all([e.correction_uncertainty_per_atom == 0 for e in entries.values()])


def test_is_experimental(entries):
    assert all([e.is_experimental for e in entries.values()])


def test_as_dict(entries):
    d = entries[300].as_dict()

    assert type(d) == dict
    assert d.get("composition")
    assert d.get("temperature")


def test_from_dict(entries):
    e = entries[300]
    d = e.as_dict()
    assert NISTReferenceEntry.from_dict(d) == e



"""
Firetasks for storing enumerated reaction or network data into a MongoDB.
"""
import datetime
import json
import os

from fireworks import FiretaskBase, FWAction, explicit_serialize
from maggma.stores import MongoStore
from monty.json import MontyDecoder, jsanitize
from monty.serialization import loadfn

from rxn_network.firetasks.utils import env_chk, get_logger

logger = get_logger(__name__)


@explicit_serialize
class ReactionsToDb(FiretaskBase):
    """
    Stores calculated reactions (rxns.json) and their metadata (metadata.json) in a
    MongoDB.
    """

    def run_task(self, fw_spec):
        db_file = env_chk(self.get("db_file"), fw_spec)

        d = {}
        rxns = loadfn("rxns.json")
        metadata = loadfn("metadata.json")

        d["name"] = (
            f"Reaction Enumeration (Target: "
            f"{metadata['target']}): {metadata['chemsys']}"
        )
        d["rxns"] = jsanitize(rxns, strict=True)
        d["metadata"] = jsanitize(metadata, strict=True)

        with MongoStore.from_db_file(db_file) as db:
            task_ids = sorted(db.distinct("task_id"))
            if task_ids:
                d["task_id"] = task_ids[-1] + 1
            else:
                d["task_id"] = 1

            d["last_updated"] = datetime.datetime.utcnow()
            db.update(d)


@explicit_serialize
class NetworkToDb(FiretaskBase):
    """
    Stores calculated reaction network in a MongoDB.
    """

    def run_task(self, fw_spec):
        pass



""" Utility functions used throughout the reaction-network package."""

from itertools import chain, combinations, zip_longest
from pathlib import Path


def limited_powerset(iterable, max_size):
    """
    Helper method for generating subsets ranging from singular
    length to maximum length specified by max_size.

    Args:
        iterable (list/set): all objects to consider.
        max_size (int): upper limit for size of combination subsets.

    Returns:
        list: all combination sets up to maximum size
    """
    return chain.from_iterable(
        [combinations(iterable, num_combos) for num_combos in range(1, max_size + 1)]
    )


def grouper(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def get_project_root() -> Path:
    return Path(__file__).parent.parent.parent



""" Tests for ReactionSet."""
from pathlib import Path
import pytest

from monty.serialization import loadfn
from pymatgen.core.composition import Element
from rxn_network.reactions.open import OpenComputedReaction
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.costs.softplus import Softplus

TEST_FILES_PATH = Path(__file__).parent.parent / "test_files"
RXNS_FILE = "ymno3_rxns.json.gz"


@pytest.fixture(scope="module")
def rxns():
    return loadfn(TEST_FILES_PATH / RXNS_FILE)


@pytest.fixture(scope="module")
def rxn_set(rxns):
    return ReactionSet.from_rxns(rxns)


def test_get_rxns(rxns, rxn_set):
    open_rxns = rxn_set.get_rxns(open_elem="O", chempot=0)
    assert rxn_set.get_rxns() == rxns
    assert open_rxns != rxns
    assert all([type(r) == OpenComputedReaction for r in open_rxns])
    assert all([r.chempots == {Element("O"): 0} for r in open_rxns])


def test_calculate_costs(rxns, rxn_set):
    cf = Softplus()
    assert rxn_set.calculate_costs(cf) == [cf.evaluate(r) for r in rxns]



"""
This module implements added features to the ChemicalPotentialDiagram class from
pymatgen.
"""

from typing import Dict, List, Optional

import numpy as np
from pymatgen.analysis.chempot_diagram import ChemicalPotentialDiagram as ChempotDiagram
from pymatgen.analysis.phase_diagram import PDEntry
from pymatgen.core.composition import Composition, Element
from scipy.spatial import KDTree


class ChemicalPotentialDiagram(ChempotDiagram):
    def __init__(
        self,
        entries: List[PDEntry],
        limits: Optional[Dict[Element, float]] = None,
        default_min_limit: Optional[float] = -20.0,
    ):
        """
        Args:
            entries: List of PDEntry-like objects containing a composition and
                energy. Must contain elemental references and be suitable for typical
                phase diagram construction. Entries must be within a chemical system
                of with 2+ elements
            limits: Bounds of elemental chemical potentials (min, max), which are
                used to construct the border hyperplanes used in the
                HalfSpaceIntersection algorithm; these constrain the space over which the
                domains are calculated and also determine the size of the plotted
                diagram. Any elemental limits not specified are covered in the
                default_min_limit argument
            default_min_limit (float): Default minimum chemical potential limit for
                unspecified elements within the "limits" argument. This results in
                default limits of (default_min_limit, 0)
        """
        super().__init__(
            entries=entries, limits=limits, default_min_limit=default_min_limit
        )

    def shortest_domain_distance(self, f1: str, f2: str) -> float:
        """
        Args:
            f1: chemical formula (1)
            f2: chemical formula (2)

        Returns:
            Shortest distance between domain boundaries in the full
            (hyper)dimensional space, calculated using KDTree.
        """
        pts1 = self.domains[Composition(f1).reduced_formula]
        pts2 = self.domains[Composition(f2).reduced_formula]

        tree = KDTree(pts1)

        return min(tree.query(pts2)[0])

    def shortest_elemental_domain_distances(self, f1, f2) -> float:
        """
        TODO: Use with caution; this function may not yet make sense geometrically!
        """
        pts1 = self.domains[f1]
        pts2 = self.domains[f2]
        pts1 = pts1[~np.isclose(pts1, self.default_min_limit).any(axis=1)]
        pts2 = pts2[~np.isclose(pts2, self.default_min_limit).any(axis=1)]
        num_elems = pts1.shape[1]

        mesh = np.meshgrid(pts1, pts2)
        diff = abs(mesh[0] - mesh[1])
        diff = diff.reshape(-1, num_elems)

        return diff.min(axis=0)



"""
A reaction class that builds reactions based on ComputedEntry objects under the
presence of an open entry (e.g. O2), and provides information about reaction
thermodynamics computed as changes in grand potential.
"""

from typing import Dict, List, Optional, Union

import numpy as np

from pymatgen.analysis.phase_diagram import GrandPotPDEntry
from pymatgen.core.composition import Element, Composition
from pymatgen.entries.computed_entries import ComputedEntry

from rxn_network.reactions.computed import ComputedReaction


class OpenComputedReaction(ComputedReaction):
    """
    Extends the ComputedReaction class to add support for "open" reactions,
    where the reaction energy is calculated as a change in grand potential.
    """

    def __init__(
        self,
        entries: List[ComputedEntry],
        coefficients: Union[np.ndarray, List[float]],
        chempots: Dict[Element, float],
        data: Optional[Dict] = None,
        lowest_num_errors=None,
    ):
        """

        Args:
            entries: List of ComputedEntry objects.
            coefficients: List of reaction coefficients.
            chempots: Dict of chemical potentials corresponding to open elements
            data: Optional dict of data
            lowest_num_errors: number of "errors" encountered during reaction balancing
        """
        super().__init__(
            entries=entries,
            coefficients=coefficients,
            data=data,
            lowest_num_errors=lowest_num_errors,
        )

        self.chempots = chempots
        self.open_elems = list(chempots.keys())

        grand_entries = []
        for e in entries:
            comp = e.composition.reduced_composition
            if len(comp.elements) == 1 and comp.elements[0] in self.open_elems:
                grand_entries.append(e)
            else:
                grand_entries.append(GrandPotPDEntry(e, chempots))

        self.grand_entries = grand_entries

    @classmethod
    def balance(  # type: ignore
        cls,
        reactant_entries: List[ComputedEntry],
        product_entries: List[ComputedEntry],
        chempots: Dict[Element, float] = None,
        data: Optional[Dict] = None,
    ):  # pylint: disable = W0221
        """

        Args:
            reactant_entries:
            product_entries:
            chempots:
            data:

        Returns:

        """

        reactant_comps = [e.composition.reduced_composition for e in reactant_entries]
        product_comps = [e.composition.reduced_composition for e in product_entries]
        coefficients, lowest_num_errors = cls._balance_coeffs(
            reactant_comps, product_comps
        )

        entries = list(reactant_entries) + list(product_entries)

        args = {
            "entries": entries,
            "coefficients": coefficients,
            "data": data,
            "lowest_num_errors": lowest_num_errors,
        }

        if not chempots:
            rxn = ComputedReaction(**args)  # type: ignore
        else:
            rxn = cls(chempots=chempots, **args)  # type: ignore

        return rxn

    @property
    def energy(self) -> float:
        """
        Returns (float):
            The calculated reaction energy.
        """
        calc_energies: Dict[Composition, float] = {}

        for entry in self.grand_entries:
            attr = "composition"
            if type(entry) == GrandPotPDEntry:
                attr = "original_comp"

            comp, factor = getattr(entry, attr).get_reduced_composition_and_factor()
            calc_energies[comp] = min(
                calc_energies.get(comp, float("inf")), entry.energy / factor
            )

        return sum(
            [
                amt * calc_energies[c]
                for amt, c in zip(self.coefficients, self.compositions)
            ]
        )

    @property
    def elements(self) -> List[Element]:
        """
        List of elements in the reaction
        """
        return list(
            set(el for comp in self.compositions for el in comp.elements)
            - set(self.open_elems)
        )

    @property
    def total_chemical_system(self) -> str:
        """
        Chemical system string, including open elements
        """
        return "-".join(
            sorted([str(e) for e in set(self.elements) | set(self.open_elems)])
        )

    def copy(self) -> "OpenComputedReaction":
        """
        Returns a copy of the OpenComputedReaction object.
        """
        return OpenComputedReaction(
            self.entries,
            self.coefficients,
            self.chempots,
            self.data,
            self.lowest_num_errors,
        )

    def reverse(self):
        """

        Returns a copy of reaction with reactants/products swapped

        """
        return OpenComputedReaction(
            self.entries,
            -1 * self.coefficients,
            self.chempots,
            self.data,
            self.lowest_num_errors,
        )

    def __repr__(self):
        cp = f"({','.join([f'mu_{e}={m}' for e, m in self.chempots.items()])})"
        return f"{super().__repr__()} {cp}"



import pytest
from pathlib import Path
from monty.serialization import loadfn
from rxn_network.entries.entry_set import GibbsEntrySet

TEST_FILES_PATH = Path(__file__).parent / "test_files"


def pytest_itemcollected(item):
    """Make tests names more readable in the tests output."""
    item._nodeid = (
        item._nodeid.replace(".py", "")
        .replace("tests/", "")
        .replace("test_", "")
        .replace("_", " ")
        .replace("Test", "")
        .replace("Class", " class")
        .lower()
    )
    doc = item.obj.__doc__.strip() if item.obj.__doc__ else ""
    if doc:
        item._nodeid = item._nodeid.split("::")[0] + "::" + doc


@pytest.fixture(
    params=[
        "Mn-O-Y_entries.json.gz",
    ],
    scope="session",
)
def mp_entries(request):
    mp_entries = loadfn(TEST_FILES_PATH / request.param)
    return mp_entries


@pytest.fixture(scope="session")
def gibbs_entries(mp_entries):
    entries = GibbsEntrySet.from_entries(mp_entries, temperature=1000)
    return entries


@pytest.fixture(scope="session")
def filtered_entries(gibbs_entries):
    filtered_entries = gibbs_entries.filter_by_stability(0.0)
    return filtered_entries


@pytest.fixture(scope="session")
def computed_rxn():
    """ 2 YOCl + 2 NaMnO2 + 0.5 O2 -> Y2Mn2O7 + 2 NaCl"""
    return loadfn(TEST_FILES_PATH / "computed_rxn.json.gz")



"""
Implements a class for conveniently and efficiently storing sets of ComputedReaction
objects which share entries.
"""

from functools import lru_cache
from typing import List, Optional, Union, Set, Iterable

import numpy as np
from monty.json import MSONable
from pymatgen.core import Element

from pymatgen.entries.computed_entries import ComputedEntry
from rxn_network.core.cost_function import CostFunction
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.open import OpenComputedReaction


class ReactionSet(MSONable):
    """
    A lightweight class for storing large sets of ComputedReaction objects.
    Automatically represents a set of reactions as an array of coefficients with
    a second array linking to a corresponding list of shared entries.
    """

    def __init__(
        self,
        entries: List[ComputedEntry],
        indices: Union[np.ndarray, List[List[int]]],
        coeffs: Union[np.ndarray, List[List[float]]],
        all_data: Optional[List] = None,
    ):
        """
        Args:
            entries: List of ComputedEntry objects shared by reactions
            indices: Array indexing the entry list; gets entries used by each
                reaction object
            coeffs: Array of all reaction coefficients
            all_data: Optional list of data for each reaction
        """
        self.entries = entries
        self.indices = indices
        self.coeffs = coeffs
        self.all_data = all_data if all_data else []

    @lru_cache(1)
    def get_rxns(
        self,
        open_elem: Optional[str] = None,
        chempot: float = 0.0,
    ) -> List[Union[ComputedReaction, OpenComputedReaction]]:
        """
        Returns list of ComputedReaction objects or OpenComputedReaction objects (when
        open element and chempot are specified) for the reaction set.

        Args:
            open_elem: Open element, e.g. "O2"
            chempot: Chemical potential (mu) of open element in equation: Phi = G - mu*N
        """
        rxns = []

        chempots = None

        if open_elem:
            chempots = {Element(open_elem): chempot}

        for indices, coeffs, data in zip(self.indices, self.coeffs, self.all_data):
            entries = [self.entries[i] for i in indices]
            if chempots:
                rxn = OpenComputedReaction(
                    entries=entries, coefficients=coeffs, data=data, chempots=chempots
                )
            else:
                rxn = ComputedReaction(entries=entries, coefficients=coeffs, data=data)
            rxns.append(rxn)
        return rxns

    def calculate_costs(
        self,
        cf: CostFunction,
        open_elem: Optional[Union[str, Element]] = None,
        chempot: Optional[float] = 0,
    ) -> List[float]:
        """
        Evaluate a cost function on an acquired set of reactions.

        Args:
            cf: CostFunction object, e.g. Softplus()
            open_elem: Open element, e.g. "O2"
            chempot: Chemical potential (mu) of open element in equation: Phi = G - mu*N
        """
        return [
            cf.evaluate(rxn)
            for rxn in self.get_rxns(open_elem=open_elem, chempot=chempot)
        ]

    @classmethod
    def from_rxns(
        cls,
        rxns: List[Union[ComputedReaction, OpenComputedReaction]],
        entries: Optional[Iterable[ComputedEntry]] = None,
    ) -> "ReactionSet":
        """
        Initiate a ReactionSet object from a list of reactions. Including a list of
        unique entries saves some computation time.

        Args:
            rxns: List of ComputedReaction-like objects.
            entries: Optional list of ComputedEntry objects
        """
        if not entries:
            entries = cls._get_unique_entries(rxns)

        entries = sorted(list(set(entries)), key=lambda r: r.composition)

        all_indices, all_coeffs, all_data = [], [], []
        for rxn in rxns:
            all_indices.append([entries.index(e) for e in rxn.entries])
            all_coeffs.append(list(rxn.coefficients))
            all_data.append(rxn.data)

        return cls(
            entries=entries,
            indices=all_indices,
            coeffs=all_coeffs,
            all_data=all_data,
        )

    @staticmethod
    def _get_unique_entries(rxns: List[ComputedReaction]) -> Set[ComputedEntry]:
        """Return only unique entries from reactions"""
        entries = set()
        for r in rxns:
            entries.update(r.entries)
        return entries



"""
Implements a reaction pathway solver class which efficiently solves mass balance
equations using matrix operations.
"""

from itertools import combinations
from typing import List

import numpy as np
from scipy.special import comb
from tqdm.notebook import tqdm

from rxn_network.core import CostFunction, Pathway, Solver
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.enumerators.basic import BasicEnumerator, BasicOpenEnumerator
from rxn_network.pathways.balanced import BalancedPathway
from rxn_network.pathways.utils import balance_path_arrays
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.utils import grouper


class PathwaySolver(Solver):
    BATCH_SIZE = 500000

    def __init__(
        self,
        entries: GibbsEntrySet,
        pathways: List[Pathway],
        cost_function: CostFunction,
        open_elem: str = None,
        chempot: float = None,
    ):
        """

        Args:
            entries:
            pathways:
            cost_function:
            open_elem:
            chempot:
        """

        super().__init__(entries=entries, pathways=pathways)
        self.cost_function = cost_function
        self.open_elem = open_elem
        self.chempot = chempot

    def solve(
        self,
        net_rxn: ComputedReaction,
        max_num_combos: int = 4,
        find_intermediate_rxns: bool = True,
        intermediate_rxn_energy_cutoff: float = 0.0,
        filter_interdependent: bool = True,
    ):
        """

        Args:
            net_rxn:
            max_num_combos:
            find_intermediate_rxns:
            intermediate_rxn_energy_cutoff:
            filter_interdependent:

        Returns:

        """

        entries = self.entries.entries_list
        precursors = net_rxn.reactant_entries
        targets = net_rxn.product_entries

        reactions = self.reactions.copy()
        costs = self.costs.copy()

        if not net_rxn.balanced:
            raise ValueError(
                "Net reaction must be balanceable to find all reaction pathways."
            )

        self.logger.info(f"NET RXN: {net_rxn} \n")

        if find_intermediate_rxns:
            self.logger.info("Identifying reactions between intermediates...")
            intermediate_rxns = self._find_intermediate_rxns(
                precursors, targets, intermediate_rxn_energy_cutoff
            )
            intermediate_costs = [
                self.cost_function.evaluate(r) for r in intermediate_rxns
            ]
            for r, c in zip(intermediate_rxns, intermediate_costs):
                if r not in reactions:
                    reactions.append(r)
                    costs.append(c)

        net_rxn_vector = self._build_idx_vector(net_rxn)
        if net_rxn in reactions:
            reactions.remove(net_rxn)

        paths = []
        for n in range(1, max_num_combos + 1):
            total = int(comb(len(reactions), n) / self.BATCH_SIZE) + 1
            groups = grouper(combinations(range(len(reactions)), n), self.BATCH_SIZE)

            pbar = groups
            if n >= 4:
                self.logger.info(f"Solving for balanced pathways of size {n} ...")
                pbar = tqdm(groups, total=total, desc="PathwaySolver")

            all_c_mats, all_m_mats = [], []
            for idx, combos in enumerate(pbar):
                if n >= 4:
                    pbar.set_description(
                        f"{self.BATCH_SIZE*idx}/{total*self.BATCH_SIZE}"
                    )
                comp_matrices = np.stack(
                    [
                        np.vstack([self._build_idx_vector(reactions[r]) for r in combo])
                        for combo in combos
                        if combo
                    ]
                )
                c_mats, m_mats = balance_path_arrays(comp_matrices, net_rxn_vector)
                all_c_mats.extend(c_mats)
                all_m_mats.extend(m_mats)

            for c_mat, m_mat in zip(all_c_mats, all_m_mats):
                path_rxns = []
                path_costs = []
                for rxn_mat in c_mat:
                    entries, coeffs = zip(
                        *[(entries[idx], c) for idx, c in enumerate(rxn_mat)]
                    )

                    rxn = ComputedReaction(entries=entries, coefficients=coeffs)
                    try:
                        path_rxns.append(rxn)
                        path_costs.append(costs[reactions.index(rxn)])
                    except Exception as e:
                        print(e)
                        continue

                p = BalancedPathway(
                    path_rxns, m_mat.flatten(), path_costs, balanced=True
                )
                paths.append(p)

        filtered_paths = []
        if filter_interdependent:
            precursor_comps = [p.composition for p in precursors]
            for p in paths:
                interdependent = p.contains_interdependent_rxns(precursor_comps)
                if not interdependent:
                    filtered_paths.append(p)
        else:
            filtered_paths = paths

        filtered_paths = sorted(list(set(filtered_paths)), key=lambda p: p.average_cost)
        return filtered_paths

    def _build_idx_vector(self, rxn):
        """

        Args:
            rxn:

        Returns:

        """
        indices = [e.data.get("idx") for e in rxn.entries]
        v = np.zeros(self.num_entries)
        v[indices] = rxn.coefficients
        return v

    def _find_intermediate_rxns(self, precursors, targets, energy_cutoff):
        rxns = []
        intermediates = {e for rxn in self.reactions for e in rxn.entries}
        intermediates = intermediates - set(precursors) - set(targets)
        intermediate_formulas = [e.composition.reduced_formula for e in intermediates]

        if not targets:
            be = BasicEnumerator(precursors=intermediate_formulas)
            rxns = be.enumerate(self.entries)
        else:
            for target in targets:
                self.logger.info(
                    f"Finding intermediate reactions to "
                    f"{target.composition.reduced_formula}..."
                )
                be = BasicEnumerator(
                    precursors=intermediate_formulas,
                    target=target.composition.reduced_formula,
                )
                int_rxns = be.enumerate(self.entries)
                rxns.extend(int_rxns)

                if self.open_elem:
                    boe = BasicOpenEnumerator(
                        open_phases=[self.open_elem],
                        precursors=intermediate_formulas,
                        target=target.composition.reduced_formula,
                    )

                    int_rxns_open = boe.enumerate(self.entries)
                    rxns.extend(int_rxns_open)

        rxns = ReactionSet.from_rxns(rxns, self.entries).get_rxns(
            open_elem=self.open_elem, chempot=self.chempot
        )
        rxns = list(filter(lambda x: x.energy_per_atom < energy_cutoff, rxns))
        self.logger.info(f"Found {len(rxns)} intermediate reactions!")

        return rxns



"""
Utility functions used in reaction pathway balancing.
"""

import numpy as np
from numba import njit, prange


@njit(parallel=True)
def balance_path_arrays(
    comp_matrices: np.ndarray,
    net_coeffs: np.ndarray,
    tol: float = 1e-6,
):
    """
    Fast solution for reaction multiplicities via mass balance stochiometric
    constraints. Parallelized using Numba.

    Args:
        comp_matrices: Array containing stoichiometric coefficients of all
            compositions in all reactions, for each trial combination.
        net_coeffs: Array containing stoichiometric coefficients of net reaction.
        tol: numerical tolerance for determining if a multiplicity is zero
            (reaction was removed).
    """
    shape = comp_matrices.shape
    net_coeff_filter = np.argwhere(net_coeffs != 0).flatten()
    len_net_coeff_filter = len(net_coeff_filter)
    all_multiplicities = np.zeros((shape[0], shape[1]), np.float64)
    indices = np.full(shape[0], False)

    for i in prange(shape[0]):
        correct = True
        for j in range(len_net_coeff_filter):
            idx = net_coeff_filter[j]
            if not comp_matrices[i][:, idx].any():
                correct = False
                break
        if not correct:
            continue

        comp_pinv = np.linalg.pinv(comp_matrices[i]).T
        multiplicities = comp_pinv @ net_coeffs
        solved_coeffs = comp_matrices[i].T @ multiplicities

        if (multiplicities < tol).any():
            continue
        elif not (
            np.abs(solved_coeffs - net_coeffs) <= (1e-08 + 1e-05 * np.abs(net_coeffs))
        ).all():
            continue
        all_multiplicities[i] = multiplicities
        indices[i] = True

    filtered_indices = np.argwhere(indices != 0).flatten()
    length = filtered_indices.shape[0]
    filtered_comp_matrices = np.empty((length, shape[1], shape[2]), np.float64)
    filtered_multiplicities = np.empty((length, shape[1]), np.float64)

    for i in range(length):
        idx = filtered_indices[i]
        filtered_comp_matrices[i] = comp_matrices[idx]
        filtered_multiplicities[i] = all_multiplicities[idx]

    return filtered_comp_matrices, filtered_multiplicities



""" Tests for ComputedReaction. Some tests adapted from pymatgen. """
import pytest
from pathlib import Path
from monty.serialization import loadfn

from rxn_network.reactions.computed import ComputedReaction

TEST_FILES_PATH = Path(__file__).parent.parent / "test_files"
ENTRIES_FILE = "yocl-namno2_rxn_entries.json.gz"


@pytest.fixture(scope="session")
def entries():
    return loadfn(TEST_FILES_PATH / ENTRIES_FILE)


@pytest.fixture(scope="session")
def reactants(entries):
    return [entries["YOCl"], entries["NaMnO2"], entries["O2"]]


@pytest.fixture(scope="session")
def products(entries):
    return [entries["Y2Mn2O7"], entries["NaCl"]]


@pytest.fixture(scope="session")
def pre_balanced_rxn(reactants, products):
    """Returns a simple, pre-balanced computed reaction."""
    coefficients = [-2, -2, -0.5, 1, 2]
    rxn = ComputedReaction(
        entries=reactants + products,
        coefficients=coefficients,
    )
    return rxn


@pytest.fixture(scope="session")
def auto_balanced_rxn(reactants, products):
    """Returns the same iron oxidation reaction, after automatically balancing"""
    return ComputedReaction.balance(
        reactant_entries=reactants, product_entries=products
    )


def test_energy(pre_balanced_rxn, auto_balanced_rxn):
    expected_energy = -2.701048

    assert pre_balanced_rxn.energy == pytest.approx(expected_energy)
    assert auto_balanced_rxn.energy == pytest.approx(expected_energy)


def test_energy_per_atom(pre_balanced_rxn, auto_balanced_rxn):
    expected_energy_per_atom = -0.1800700

    assert pre_balanced_rxn.energy_per_atom == pytest.approx(expected_energy_per_atom)
    assert auto_balanced_rxn.energy_per_atom == pytest.approx(expected_energy_per_atom)


def test_energy_uncertainty(pre_balanced_rxn, auto_balanced_rxn):
    expected_energy_uncertainty = 0.0229486383

    assert pre_balanced_rxn.energy_uncertainty == pytest.approx(
        expected_energy_uncertainty
    )
    assert auto_balanced_rxn.energy_uncertainty == pytest.approx(
        expected_energy_uncertainty
    )


def test_energy_uncertainty_per_atom(pre_balanced_rxn, auto_balanced_rxn):
    expected_energy_uncertainty_per_atom = 0.0015299092

    assert pre_balanced_rxn.energy_uncertainty_per_atom == pytest.approx(
        expected_energy_uncertainty_per_atom
    )
    assert auto_balanced_rxn.energy_uncertainty_per_atom == pytest.approx(
        expected_energy_uncertainty_per_atom
    )


def test_copy(pre_balanced_rxn, auto_balanced_rxn):
    pre_balanced_rxn_copy = pre_balanced_rxn.copy()
    auto_balanced_rxn_copy = auto_balanced_rxn.copy()

    assert (
        pre_balanced_rxn
        == auto_balanced_rxn
        == pre_balanced_rxn_copy
        == auto_balanced_rxn_copy
    )


def test_reverse(pre_balanced_rxn):
    pre_balanced_rxn_rev = pre_balanced_rxn.reverse()

    assert pre_balanced_rxn.energy == -pre_balanced_rxn_rev.energy

    assert pre_balanced_rxn == pre_balanced_rxn_rev.reverse()



"""
Firetasks for running enumeration and network calculations
"""
import json

from fireworks import FiretaskBase, FWAction, explicit_serialize
from monty.json import MontyEncoder
from monty.serialization import dumpfn
from pymatgen.core import Composition

from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.firetasks.utils import env_chk, get_logger
from rxn_network.reactions.reaction_set import ReactionSet

logger = get_logger(__name__)


@explicit_serialize
class RunEnumerators(FiretaskBase):
    """
    Run a list of enumerators on a provided set of computed entries and dump the
    calculated ComputedReaction objects to a file (rxns.json). Metadata is stored as metadata.json.

    Required params:
        enumerators (List[Enumerator]): Enumerators to run
        entries (List[ComputedEntry]): Computed entries to be fed into enumerate()
            methods

    """

    required_params = ["enumerators", "entries"]

    def run_task(self, fw_spec):
        enumerators = self["enumerators"]
        entries = self.get("entries", None)

        if not entries:
            entries = fw_spec["entries"]
        else:
            entries = entries["entries"]

        entries = GibbsEntrySet(entries)
        chemsys = "-".join(sorted(list(entries.chemsys)))
        target = enumerators[0].target
        added_elems = None

        if target:
            added_elems = entries.chemsys - {
                str(e) for e in Composition(target).elements
            }
            added_elems = "-".join(sorted(list(added_elems)))

        metadata = {
            "chemsys": chemsys,
            "enumerators": enumerators,
            "target": target,
            "added_elems": added_elems,
        }

        results = []
        for enumerator in enumerators:
            rxns = enumerator.enumerate(entries)
            results.extend(rxns)

        results = ReactionSet.from_rxns(results)

        dumpfn(results, "rxns.json")
        dumpfn(metadata, "metadata.json")


@explicit_serialize
class RunNetwork(FiretaskBase):
    pass



"""
Utility functions used in the ReactionNetwork class.
"""

from rxn_network.network.entry import NetworkEntry, NetworkEntryType


def get_rxn_nodes_and_edges(rxns):
    """

    Args:
        rxns:

    Returns:

    """
    nodes, edges = [], []

    for rxn in rxns:
        reactant_node = NetworkEntry(rxn.reactant_entries, NetworkEntryType.Reactants)
        product_node = NetworkEntry(rxn.product_entries, NetworkEntryType.Products)

        if reactant_node not in nodes:
            nodes.append(reactant_node)
            reactant_idx = len(nodes) - 1
        else:
            reactant_idx = nodes.index(reactant_node)

        if product_node not in nodes:
            nodes.append(product_node)
            product_idx = len(nodes) - 1
        else:
            product_idx = nodes.index(product_node)

        edges.append((reactant_idx, product_idx))

    return nodes, edges


def get_loopback_edges(g, nodes):
    """

    Args:
        g:
        nodes:

    Returns:

    """
    edges = []
    for idx1, p in enumerate(nodes):
        if p.description.value != NetworkEntryType.Products.value:
            continue
        for idx2, r in enumerate(nodes):
            if r.description.value != NetworkEntryType.Reactants.value:
                continue
            if p.entries == r.entries:
                edges.append((g.vertex(idx1), g.vertex(idx2), 0.0, None, "loopback"))

    return edges



"""
Helpful utility functions used by the enumerator classes.
"""
from typing import List, Union

from pymatgen.entries.computed_entries import Entry, ComputedEntry

import rxn_network.costs.calculators as calcs
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.open import OpenComputedReaction


def initialize_entry(formula: str, entry_set: GibbsEntrySet, stabilize: bool = True):
    """
    Acquire a (stabilized) entry by user-specified formula.

    Args:
        formula: Chemical formula
        entry_set: GibbsEntrySet containing 1 or more entries corresponding to
            given formula
        stabilize: Whether or not to stabilize the entry by decreasing its energy
            such that it is 'on the hull'
    """
    entry = entry_set.get_min_entry_by_formula(formula)

    if stabilize:
        entry = entry_set.stabilize_entry(entry)
    return entry


def initialize_calculators(
    calculators: Union[List[calcs.Calculator], List[str]], entries: GibbsEntrySet
):
    """
    Initialize a list of Calculators given a list of their names (strings) or
    uninitialized objects, and a provided list of entries.

    Args:
        calculators: List of names of calculators
        entries: List of entries or EntrySet-type object
    """
    calculators = [getattr(calcs, c) if isinstance(c, str) else c for c in calculators]
    return [c.from_entries(entries) for c in calculators]  # type: ignore


def apply_calculators(rxn: ComputedReaction, calculators: List[calcs.Calculator]):
    """
    Decorates a reaction by applying decorate() from a list of calculators.

    Args:
        rxn: ComputedReaction object
        calculators: List of (initialized) calculators

    """
    for calc in calculators:
        rxn = calc.decorate(rxn)
    return rxn


def get_total_chemsys(entries: List[Entry], open_elem=None):
    """
    Returns chemical system for set of entries, with optional open element.

    Args:
        entries:
        open_elem:
    """
    elements = {elem for entry in entries for elem in entry.composition.elements}
    if open_elem:
        elements.add(open_elem)
    return "-".join(sorted([str(e) for e in elements]))


def get_elems_set(entries):
    """

    Args:
        entries:

    Returns:

    """
    return {str(elem) for e in entries for elem in e.composition.elements}


def group_by_chemsys(combos, open_elem=None):
    """

    Args:
        combos:
        open_elem:

    Returns:

    """
    combo_dict = {}
    for combo in combos:
        key = get_total_chemsys(combo, open_elem)
        if key in combo_dict:
            combo_dict[key].append(combo)
        else:
            combo_dict[key] = [combo]

    return combo_dict


def stabilize_entries(pd, entries_to_adjust, tol=1e-6):
    """

    Args:
        pd:
        entries_to_adjust:
        tol:

    Returns:

    """
    indices = [pd.all_entries.index(entry) for entry in entries_to_adjust]
    new_entries = []
    for idx, entry in zip(indices, entries_to_adjust):
        e_above_hull = pd.get_e_above_hull(entry)
        entry_dict = entry.to_dict()
        entry_dict["energy"] = entry.uncorrected_energy + (
            e_above_hull * entry.composition.num_atoms - tol
        )
        new_entry = ComputedEntry.from_dict(entry_dict)
        new_entries.append(new_entry)
    return new_entries


def filter_entries_by_chemsys(entries, chemsys):
    """

    Args:
        entries:
        chemsys:

    Returns:

    """
    chemsys = set(chemsys.split("-"))

    filtered_entries = list(
        filter(
            lambda e: chemsys.issuperset(e.composition.chemical_system.split("-")),
            entries,
        )
    )
    return filtered_entries


def get_entry_by_comp(comp, entries):
    """

    Args:
        comp:
        entries:

    Returns:

    """
    possible_entries = filter(
        lambda e: e.composition.reduced_composition == comp.reduced_composition, entries
    )
    return sorted(possible_entries, key=lambda e: e.energy_per_atom)[0]


def get_computed_rxn(rxn, entries):
    """

    Args:
        rxn:
        entries:

    Returns:

    """
    reactant_entries = [get_entry_by_comp(r, entries) for r in rxn.reactants]
    product_entries = [get_entry_by_comp(p, entries) for p in rxn.products]

    rxn = ComputedReaction.balance(reactant_entries, product_entries)
    return rxn


def get_open_computed_rxn(rxn, entries, chempots):
    """

    Args:
        rxn:
        entries:
        chempots:

    Returns:

    """
    reactant_entries = [get_entry_by_comp(r, entries) for r in rxn.reactants]
    product_entries = [get_entry_by_comp(p, entries) for p in rxn.products]

    rxn = OpenComputedReaction.balance(reactant_entries, product_entries, chempots)

    return rxn



""" Tests for GibbsComputedEntry. Some tests adapted from pymatgen."""
from pathlib import Path

import pytest
from monty.serialization import loadfn

from rxn_network.entries.gibbs import GibbsComputedEntry

TEST_FILES_PATH = Path(__file__).parent.parent / "test_files"


@pytest.fixture
def structure():
    struct = loadfn(TEST_FILES_PATH / "structure_LiFe4P4O16.json")
    return struct


@pytest.fixture
def entry(structure):
    entry = GibbsComputedEntry.from_structure(
        structure=structure,
        formation_energy_per_atom=-2.436,
        temperature=300,
        parameters=None,
        entry_id="LiFe4P4O16 test structure",
    )
    return entry


@pytest.fixture
def entries_temps_dict(structure):
    struct = structure

    temps = [300, 600, 900, 1200, 1500, 1800]
    entries_with_temps = {
        temp: GibbsComputedEntry.from_structure(
            structure=struct,
            formation_energy_per_atom=-2.436,
            temperature=temp,
            parameters=None,
            entry_id="Test LiFe4P4O16 structure",
        )
        for temp in temps
    }
    return entries_with_temps


def test_gf_sisso(entries_temps_dict):
    test_energies = {
        300: -56.21273010866969,
        600: -51.52997063074788,
        900: -47.29888391585979,
        1200: -42.942338738866304,
        1500: -37.793417248809774,
        1800: -32.32513382051749,
    }
    entry_energies = {t: e.energy for t, e in entries_temps_dict.items()}

    assert entry_energies == pytest.approx(test_energies)


def test_interpolation(structure):
    temp = 450
    e = GibbsComputedEntry.from_structure(
        structure=structure, formation_energy_per_atom=-2.436, temperature=temp
    )
    assert e.energy == pytest.approx(-53.7243542548528)


def test_to_from_dict(entry):
    d = entry.as_dict()
    e = GibbsComputedEntry.from_dict(d)
    assert e == entry
    assert e.energy == pytest.approx(entry.energy)


def test_str(entry):
    assert str(entry) is not None


def test_normalize(entries_temps_dict):
    num_atoms = 25
    for e in entries_temps_dict.values():
        normed_entry = e.normalize(mode="atom")
        assert e.uncorrected_energy == pytest.approx(
            normed_entry.uncorrected_energy * num_atoms
        )



""" Tests for ChempotDistanceCalculator """
from pathlib import Path

import pytest
from monty.serialization import loadfn

import numpy as np

from rxn_network.costs.calculators import ChempotDistanceCalculator
from rxn_network.thermo.chempot_diagram import ChemicalPotentialDiagram
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.reactions.computed import ComputedReaction

TEST_FILES_PATH = Path(__file__).parent.parent / "test_files"

answers = {
    "0.5 Y2O3 + 0.5 Mn2O3 -> YMnO3": {
        np.sum: 0.480008216,
        np.max: 0.480008216,
        np.mean: 0.240004108,
    },
    "2 YClO + 2 NaMnO2 + 0.5 O2 -> Y2Mn2O7 + 2 NaCl": {
        np.sum: 1.369790046,
        np.max: 1.369790045,
        np.mean: 0.195684292,
    },
}


@pytest.fixture(params=["Cl-Mn-Na-O-Y_entries.json.gz"])
def entries(request):
    entries = loadfn(TEST_FILES_PATH / request.param)
    return GibbsEntrySet(entries)


@pytest.fixture(
    params=[
        [["Y2O3", "Mn2O3"], ["YMnO3"]],
        [["YOCl", "NaMnO2", "O2"], ["Y2Mn2O7", "NaCl"]],
    ]
)
def rxn(entries, request):
    reactants = request.param[0]
    products = request.param[1]
    reactant_entries = [entries.get_min_entry_by_formula(r) for r in reactants]
    product_entries = [entries.get_min_entry_by_formula(p) for p in products]
    return ComputedReaction.balance(reactant_entries, product_entries)


@pytest.fixture(params=["sum", "max", "mean"])
def mu_func(request):
    return request.param


@pytest.fixture
def cpd(entries):
    return ChemicalPotentialDiagram(entries)


@pytest.fixture
def calculator(cpd, mu_func):
    return ChempotDistanceCalculator(cpd=cpd, mu_func=mu_func)


def test_calculate(calculator, rxn):
    actual_cost = calculator.calculate(rxn)
    expected_cost = answers[str(rxn)][calculator.mu_func]

    assert actual_cost == pytest.approx(expected_cost)


def test_decorate(calculator, rxn):
    rxn_dec = calculator.decorate(rxn)

    actual_cost = rxn_dec.data[calculator.name]
    expected_cost = answers[str(rxn)][calculator.mu_func]

    assert type(rxn_dec) == ComputedReaction
    assert pytest.approx(expected_cost, actual_cost)


def test_from_entries(entries, mu_func, rxn):
    calc = ChempotDistanceCalculator.from_entries(entries=entries, mu_func=mu_func)

    actual_cost = calc.calculate(rxn)
    expected_cost = answers[str(rxn)][calc.mu_func]

    assert type(calc) == ChempotDistanceCalculator
    assert pytest.approx(expected_cost, actual_cost)



""" Tests for MinimizeGibbsEnumerator and MinimizeGrandPotentialEnumerator """
import pytest
from pymatgen.core.composition import Element
from rxn_network.enumerators.minimize import (
    MinimizeGibbsEnumerator,
    MinimizeGrandPotentialEnumerator,
)


@pytest.fixture
def gibbs_enumerator_default():
    return MinimizeGibbsEnumerator()


@pytest.fixture
def gibbs_enumerator_with_calculator():
    return MinimizeGibbsEnumerator(calculators=["ChempotDistanceCalculator"])


@pytest.fixture
def gibbs_enumerator_with_precursors():
    return MinimizeGibbsEnumerator(precursors=["Y2O3", "Mn2O3"])


@pytest.fixture
def gibbs_enumerator_with_target():
    return MinimizeGibbsEnumerator(target="YMnO3")


@pytest.fixture
def gibbs_enumerator_with_precursors_and_target():
    return MinimizeGibbsEnumerator(precursors=["Y2O3", "Mn2O3"], target="YMnO3")


@pytest.fixture
def grand_potential_enumerator():
    return MinimizeGrandPotentialEnumerator(open_elem=Element("O"), mu=0.0)


@pytest.fixture
def grand_potential_enumerator_with_precursors():
    return MinimizeGrandPotentialEnumerator(
        open_elem=Element("O"), mu=0.0, precursors=["Y2O3", "Mn2O3"]
    )


@pytest.fixture
def grand_potential_enumerator_with_target():
    return MinimizeGrandPotentialEnumerator(
        open_elem=Element("O"), mu=0.0, target="Y2Mn2O7"
    )


@pytest.fixture
def grand_potential_enumerator_with_precursors_and_target():
    return MinimizeGrandPotentialEnumerator(
        open_elem=Element("O"), mu=0.0, precursors=["Y2O3", "Mn2O3"], target="Y2Mn2O7"
    )


def test_enumerate_gibbs(
    filtered_entries, gibbs_enumerator_default, gibbs_enumerator_with_calculator
):
    expected_num_rxns = 109

    for enumerator in [gibbs_enumerator_default, gibbs_enumerator_with_calculator]:
        rxns = enumerator.enumerate(filtered_entries)

        assert len(rxns) == expected_num_rxns
        assert len(rxns) == len(set(rxns))
        assert all([not r.is_identity for r in rxns])

        if enumerator.calculators:
            assert all([r.data["chempot_distance"] is not None for r in rxns])


def test_enumerate_gibbs_with_precursors(
    filtered_entries, gibbs_enumerator_with_precursors
):
    expected_num_rxns = 2

    rxns = gibbs_enumerator_with_precursors.enumerate(filtered_entries)
    precursors = gibbs_enumerator_with_precursors.precursors

    assert len(rxns) == expected_num_rxns

    for r in rxns:
        reactants = [i.reduced_formula for i in r.reactants]
        products = [i.reduced_formula for i in r.products]

        for precursor in precursors:
            assert precursor in reactants
            assert precursor not in products


def test_enumerate_gibbs_with_target(filtered_entries, gibbs_enumerator_with_target):
    expected_num_rxns = 32

    rxns = gibbs_enumerator_with_target.enumerate(filtered_entries)
    target = gibbs_enumerator_with_target.target

    assert len(rxns) == expected_num_rxns

    for r in rxns:
        reactants = [i.reduced_formula for i in r.reactants]
        products = [i.reduced_formula for i in r.products]
        assert target not in reactants
        assert target in products


def test_enumerate_gibbs_with_precursors_and_target(
    filtered_entries, gibbs_enumerator_with_precursors_and_target
):
    rxns = gibbs_enumerator_with_precursors_and_target.enumerate(filtered_entries)

    assert {str(r) for r in rxns} == {"0.5 Mn2O3 + 0.5 Y2O3 -> YMnO3"}


def test_enumerate_grand_potential(filtered_entries, grand_potential_enumerator):
    expected_num_rxns = 39

    rxns = grand_potential_enumerator.enumerate(filtered_entries)

    assert len(rxns) == expected_num_rxns
    assert len(rxns) == len(set(rxns))
    assert all([not r.is_identity for r in rxns])


def test_enumerate_grand_potential_precursors(
    filtered_entries, grand_potential_enumerator_with_precursors
):
    expected_rxns = {
        "Mn2O3 + 0.5 Y2O3 + 0.25 O2 -> YMn2O5",
        "Mn2O3 + Y2O3 + 0.5 O2 -> Y2Mn2O7",
    }

    rxns = grand_potential_enumerator_with_precursors.enumerate(filtered_entries)
    assert {str(r) for r in rxns} == expected_rxns


def test_enumerate_grand_potential_target(
    filtered_entries, grand_potential_enumerator_with_target
):
    expected_num_rxns = 12

    rxns = grand_potential_enumerator_with_target.enumerate(filtered_entries)
    target = grand_potential_enumerator_with_target.target

    assert len(rxns) == expected_num_rxns

    for r in rxns:
        reactants = [i.reduced_formula for i in r.reactants]
        products = [i.reduced_formula for i in r.products]
        assert target not in reactants
        assert target in products


def test_enumerate_grand_potential_precursors_target(
    filtered_entries, grand_potential_enumerator_with_precursors_and_target
):
    rxns = grand_potential_enumerator_with_precursors_and_target.enumerate(
        filtered_entries
    )

    assert len(rxns) == 1
    assert str(rxns[0]) == "Mn2O3 + Y2O3 + 0.5 O2 -> Y2Mn2O7"



"""
Implements an Entry that looks up NIST pre-tabulated Gibbs free energies
"""
import hashlib
from typing import Dict, Any

from pymatgen.core.composition import Composition
from pymatgen.entries import Entry
from scipy.interpolate import interp1d

from rxn_network.data import G_COMPOUNDS, G_GASES


class NISTReferenceEntry(Entry):
    """
    An Entry class for NIST-JANAF experimental reference data. Given a composition,
    automatically finds the Gibbs free energy of formation, dGf(T) from tabulated
    reference values (G_GASES, G_COMPOUNDS).

    Reference:
        Malcolm W. Chase Jr. NIST-JANAF thermochemical tables. Fourth edition.
        Washington, DC : American Chemical Society;  New York : American Institute of
        Physics for the National Institute of Standards and Technology, 1998.
    """

    REFERENCES = {**G_COMPOUNDS, **G_GASES}

    def __init__(self, composition: Composition, temperature: float):
        """
        Args:
            composition: Composition object (pymatgen).
            temperature: Temperature in Kelvin. If temperature is not selected from
                one of [300, 400, 500, ... 2000 K], then free energies will be
                interpolated. Defaults to 300 K.
        """
        composition = Composition(composition)
        formula = composition.reduced_formula

        if formula not in NISTReferenceEntry.REFERENCES:
            raise ValueError("Formula must be in NIST-JANAF thermochemical tables")

        if temperature < 300 or temperature > 2000:
            raise ValueError("Temperature must be selected from range: [300, 2000] K")

        energy = self._get_nist_energy(formula, temperature)

        self.temperature = temperature
        self._formula = formula
        self.name = formula
        self.entry_id = "NISTReferenceEntry"
        self.data = {}  # type: Dict[Any, Any]

        super().__init__(composition.reduced_composition, energy)

    @staticmethod
    def _get_nist_energy(formula: str, temperature: float) -> float:
        """
        Convenience method for accessing and interpolating NIST-JANAF data.

        Args:
            formula: Chemical formula by which to search NIST-JANAF data.
            temperature: Absolute temperature [K].

        Returns:
            Gibbs free energy of formation of formula at specified temperature [eV]
        """
        data = NISTReferenceEntry.REFERENCES[formula]
        if temperature % 100 > 0:
            g_interp = interp1d([float(t) for t in data.keys()], list(data.values()))
            return g_interp(temperature)[()]

        return data[str(temperature)]

    @property
    def energy(self) -> float:
        """The energy of the entry, as supplied by the NIST-JANAF tables."""
        return self._energy

    @property
    def correction_uncertainty(self) -> float:
        """ Uncertainty of NIST-JANAF data is not supplied."""
        return 0

    @property
    def correction_uncertainty_per_atom(self) -> float:
        """Uncertainty of NIST-JANAF data is not supplied."""
        return 0

    @property
    def is_experimental(self) -> bool:
        """ Returns True by default."""
        return True

    def as_dict(self) -> dict:
        """ Returns an MSONable dict."""
        data = super().as_dict()
        data["temperature"] = self.temperature
        return data

    @classmethod
    def from_dict(cls, d) -> "NISTReferenceEntry":
        """ Returns NISTReferenceEntry constructed from MSONable dict."""
        return cls(composition=d["composition"], temperature=d["temperature"])

    def __repr__(self):
        output = [
            f"NISTReferenceEntry | {self._formula}",
            f"Gibbs Energy ({self.temperature} K) = {self.energy:.4f}",
        ]
        return "\n".join(output)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self._formula == other._formula) and (
                self.temperature == other.temperature
            )
        return False

    def __hash__(self):
        data_md5 = hashlib.md5(
            "NISTReferenceEntry"
            f"{self.composition}_"
            f"{self.temperature}".encode("utf-8")
        ).hexdigest()
        return int(data_md5, 16)



"""
Test for BasicReaction. Several tests adapted from
test module for pymatgen.analysis.reaction_calculator
"""
import pytest
from pymatgen.core.composition import Composition, Element

from rxn_network.reactions.basic import BasicReaction


@pytest.fixture
def pre_balanced_rxn():
    """Returns a simple, pre-balanced iron oxidation reaction."""
    reactants = [Composition("Fe"), Composition("O2")]
    products = [Composition("Fe2O3")]
    coefficients = [-2, -1.5, 1]

    rxn = BasicReaction(
        compositions=reactants + products, coefficients=coefficients, balanced=True
    )
    return rxn


@pytest.fixture
def auto_balanced_rxn():
    """Returns the same iron oxidation reaction, after automatically balancing"""
    reactants = ["Fe", "O2"]
    products = ["Fe2O3"]
    return BasicReaction.from_formulas(reactants, products)


@pytest.mark.parametrize(
    "reactants, products, expected_rxn, expected_lowest_num_errors",
    [
        (["MgO"], ["MgO"], "MgO -> MgO", 0),
        (["Fe", "O2"], ["Fe2O3"], "2 Fe + 1.5 O2 -> Fe2O3", 0),
        (["FePO4", "LiPO3"], ["LiFeP2O7"], "FePO4 + LiPO3 -> LiFeP2O7", 0),
        (["Zn", "HCl"], ["ZnCl2", "H2"], "Zn + 2 HCl -> ZnCl2 + H2", 0),
        (
            ["LiCrO2", "La8Ti8O12", "O2"],
            ["LiLa3Ti3CrO12"],
            "LiCrO2 + 1.5 La2Ti2O3 + 2.75 O2 -> LiLa3Ti3CrO12",
            0,
        ),
        (["Na", "K2O"], ["Na2O", "K"], "2 Na + K2O -> Na2O + 2 K", 0),
        (
            ["NaMnO2", "YClO", "O2"],
            ["Y2Mn2O7", "NaCl"],
            "2 NaMnO2 + 2 YClO + 0.5 O2 -> Y2Mn2O7 + 2 NaCl",
            0,
        ),
        (
            ["La2O3", "Co2O3", "Li2ZrO3"],
            ["Li2O", "La2Zr2O7", "Li3CoO3"],
            "La2O3 + 0.3333 Co2O3 + 2 Li2ZrO3 -> Li2O + La2Zr2O7 + 0.6667 Li3CoO3",
            0,
        ),
        (["Li", "Cl", "Cl"], ["LiCl"], "Li + 0.25 Cl2 + 0.25 Cl2 -> LiCl", 0),
        (
            ["LiMnCl3", "LiCl", "MnCl2"],
            ["Li2MnCl4"],
            "LiMnCl3 + 3 LiCl + MnCl2 -> 2 Li2MnCl4",
            0,
        ),
        (["Fe", "O2"], ["Fe", "O2"], "Fe + O2 -> Fe + O2", 0),
        (
            ["Fe", "O2", "Na", "Li", "Cl"],
            ["FeO2", "NaCl", "Li2Cl2"],
            "Fe + O2 + Na + 2 Li + 1.5 Cl2 -> FeO2 + NaCl + 2 LiCl",
            0,
        ),
        (["FePO4", "O"], ["FePO4"], "FePO4 -> FePO4", 1),
        (
            ["La2O3", "Co2O3", "Li2ZrO3"],
            ["Li2O", "La2Zr2O7", "Li3CoO3", "Xe"],
            "La2O3 + 0.3333 Co2O3 + 2 Li2ZrO3 -> Li2O + La2Zr2O7 + 0.6667 Li3CoO3",
            1,
        ),
        (["FePO4", "Mn"], ["FePO4", "Xe"], "FePO4 -> FePO4", 2),
        (
            ["La2O3", "Co2O3", "Li2ZrO3"],
            ["Li2O", "La2Zr2O7", "Li3CoO3", "Xe", "XeNe"],
            "La2O3 + 0.3333 Co2O3 + 2 Li2ZrO3 -> Li2O + La2Zr2O7 + 0.6667 Li3CoO3",
            2,
        ),
        (
            ["LiCoO2"],
            ["La2O3", "Co2O3", "Li2O", "LiF", "CoF3"],
            "1.667 LiCoO2 + 0.3333 CoF3 -> Co2O3 + 0.3333 Li2O + LiF",
            2,
        ),
        (["LiCoO2", "Li2O"], ["ZrF4", "Co2O3"], "2 LiCoO2 -> Li2O + Co2O3", 2),
        (
            ["Fe", "Na", "Li2O", "Cl"],
            ["LiCl", "Na2O", "Xe", "FeCl", "Mn"],
            "Fe + Na + 0.5 Li2O + Cl2 -> LiCl + 0.5 Na2O + FeCl",
            2,
        ),
        (["XeMn", "Li"], ["S", "LiS2", "FeCl"], "Li + 2 S -> LiS2", 3),
    ],
)
def test_balance(reactants, products, expected_rxn, expected_lowest_num_errors):
    rxn = BasicReaction.from_formulas(reactants, products)
    assert str(rxn) == expected_rxn
    assert rxn.lowest_num_errors == expected_lowest_num_errors


def test_equality(pre_balanced_rxn, auto_balanced_rxn):
    assert pre_balanced_rxn == auto_balanced_rxn


@pytest.mark.parametrize(
    "reactants, products",
    [
        (["YMnO3"], ["YMnO3"]),
        (["YMnO3", "O2"], ["YMnO3", "O2"]),
        (["Li", "Na", "K"], ["Li", "Na", "K"]),
    ],
)
def test_is_identity(reactants, products):
    rxn = BasicReaction.from_formulas(reactants, products)

    assert rxn.is_identity is True


def test_get_el_amount(pre_balanced_rxn):
    rxn_normalized_oxygen = pre_balanced_rxn.copy().normalize_to_element(Element("O"))
    assert pre_balanced_rxn.get_el_amount(Element("Fe")) == 2
    assert pre_balanced_rxn.get_el_amount(Element("O")) == 3
    assert rxn_normalized_oxygen.get_el_amount(Element("O")) == 1


def test_copy(pre_balanced_rxn):
    rxn_copy = pre_balanced_rxn.copy()
    assert rxn_copy == pre_balanced_rxn.copy()


def test_equals():
    rxn1 = BasicReaction.from_formulas(["Y2O3", "MnO2"], ["Y2Mn2O7"])
    rxn2 = BasicReaction.from_formulas(["MnO2", "Y2O3"], ["Y2Mn2O7"])

    assert rxn1 == rxn2


def test_reverse():
    rxn = BasicReaction.from_formulas(
        ["La2O3", "Co2O3", "Li2ZrO3"], ["Li2O", "La2Zr2O7", "Li3CoO3", "Xe"]
    )
    rxn_reverse = rxn.reverse()

    assert (
        rxn_reverse
        == "Li2O + La2Zr2O7 + 0.6667 Li3CoO3 -> La2O3 + 0.3333 Co2O3 + 2 Li2ZrO3"
    )
    assert rxn_reverse.reverse() == rxn


def test_normalize(pre_balanced_rxn):
    rxn, factor = pre_balanced_rxn.normalized_repr_and_factor()
    rxn_o2_norm = pre_balanced_rxn.normalize_to(Composition("O2"))
    rxn_o2_norm_5 = pre_balanced_rxn.normalize_to(Composition("O2"), 5)

    assert rxn == "4 Fe + 3 O2 -> 2 Fe2O3"
    assert factor == 2
    assert rxn_o2_norm == "1.333 Fe + O2 -> 0.6667 Fe2O3"
    assert rxn_o2_norm_5 == "6.667 Fe + 5 O2 -> 3.333 Fe2O3"


def test_from_string(pre_balanced_rxn):
    prebalanced_rxn_from_str = BasicReaction.from_string("2 Fe + 1.5 O2 -> Fe2O3")

    rxn = BasicReaction.from_formulas(["Y2O3", "MnO2"], ["Y2Mn2O7"])
    rxn_from_str = BasicReaction.from_string("Y2O3 + 2 MnO2 -> Y2Mn2O7")

    assert pre_balanced_rxn == prebalanced_rxn_from_str
    assert rxn == rxn_from_str


def test_chemical_system(pre_balanced_rxn):
    complex_rxn = BasicReaction.from_formulas(
        ["LiCoO2", "Be", "Na"], ["La2O3", "Co2O3", "Li2O", "LiF", "CoF3"]
    )
    assert pre_balanced_rxn.chemical_system == "Fe-O"
    assert complex_rxn.chemical_system == "Be-Co-F-La-Li-Na-O"


def test_balanced():
    rxn_unbalanced = BasicReaction.from_formulas(["MnO2", "Y2O3"], ["YMn2O5"])
    rxn_balanced = BasicReaction.from_formulas(["YClO", "LiMnO2"], ["YMnO3", "LiCl"])
    rxn_balanced_2 = BasicReaction.from_formulas(
        ["Li", "Na2O", "FeCl2", "Y2O3"], ["YCl3", "Fe2O3", "NaCl", "Li2O"]
    )

    assert rxn_unbalanced.balanced is False
    assert rxn_balanced.balanced is True
    assert rxn_balanced_2.balanced is True


def test_get_coeff(pre_balanced_rxn):
    expected_coeffs = [-2, -1.5, 1]
    coeffs = [pre_balanced_rxn.get_coeff(c) for c in pre_balanced_rxn.compositions]

    assert coeffs == expected_coeffs


def test_energy(pre_balanced_rxn):
    with pytest.raises(ValueError) as error:
        assert pre_balanced_rxn.energy
    assert str(error.value) == "No energy for a basic reaction!"


def test_energy_per_atom(pre_balanced_rxn):
    with pytest.raises(ValueError) as error:
        assert pre_balanced_rxn.energy_per_atom
    assert str(error.value) == "No energy per atom for a basic reaction!"


def test_reactants_products(pre_balanced_rxn):
    assert pre_balanced_rxn.reactants == [Composition("Fe"), Composition("O2")]
    assert pre_balanced_rxn.products == [Composition("Fe2O3")]


def test_to_from_dict(pre_balanced_rxn):
    d = pre_balanced_rxn.as_dict()
    rxn = BasicReaction.from_dict(d)

    assert rxn == pre_balanced_rxn


def test_scientific_notation():
    reactants = ["FePO4"]
    products = ["FePO3.9999", "O2"]
    rxn = BasicReaction.from_formulas(reactants, products)
    rxn2 = BasicReaction.from_string("1e3 FePO4 + 20 CO -> 1e1 O2 + 1e3 Fe1P1O4 + 20 C")

    assert str(rxn) == "FePO4 -> Fe1P1O3.9999 + 5e-05 O2"
    assert str(rxn2) == "1000 FePO4 + 20 CO -> 10 O2 + 1000 FePO4 + 20 C"



""" Implementations of Reaction Network Pathway """
from rxn_network.pathways.balanced import BalancedPathway
from rxn_network.pathways.basic import BasicPathway
from rxn_network.pathways.solver import PathwaySolver



"""
Basic interface for a reaction cost Calculator
"""
from abc import ABCMeta, abstractmethod

from monty.json import MSONable

from rxn_network.core.reaction import Reaction


class Calculator(MSONable, metaclass=ABCMeta):
    """
    Base definition for a property calculator
    """

    @abstractmethod
    def calculate(self, rxn: Reaction) -> float:
        """
        Evaluates the specified property of a reaction
        """

    @abstractmethod
    def decorate(self, rxn: Reaction) -> "Reaction":
        """
        Evaluates the specified prop. of a reaction and stores it in the reaction data
        """



from itertools import combinations, product
from math import comb
from typing import List, Optional

from pymatgen.analysis.interface_reactions import InterfacialReactivity
from pymatgen.core.composition import Element
from pymatgen.entries.computed_entries import ComputedEntry
from tqdm.auto import tqdm

from rxn_network.enumerators.basic import BasicEnumerator
from rxn_network.enumerators.utils import (
    apply_calculators,
    get_computed_rxn,
    get_open_computed_rxn,
)


class MinimizeGibbsEnumerator(BasicEnumerator):
    """
    Enumerator for finding all reactions between two reactants that are predicted by
    thermodynamics; i.e., they appear when taking the convex hull along a straight
    line connecting any two phases in G-x phase space. Identity reactions are
    automatically excluded.
    """

    def __init__(
        self,
        precursors: Optional[List[str]] = None,
        target: Optional[str] = None,
        calculators: Optional[List[str]] = None,
    ):
        """
        Args:
            precursors: Optional formulas of precursors.
            target: Optional formula of target; only reactions which make this target
                will be enumerated.
            calculators: Optional list of Calculator object names; see calculators
                module for options (e.g., ["ChempotDistanceCalculator])
        """
        super().__init__(precursors, target, calculators)
        self._build_pd = True

    def estimate_max_num_reactions(self, entries: List[ComputedEntry]) -> int:
        """
        Estimate the upper bound of the number of possible reactions. This will
        correlate with the amount of time it takes to enumerate reactions.

        Args:
            entries: A list of all entries to consider

        Returns: The upper bound on the number of possible reactions
        """
        return comb(len(entries), 2) * 2

    def _react(self, reactants, products, calculators, pd=None, grand_pd=None):
        r = list(reactants)
        r0 = r[0]

        if len(r) == 1:
            r1 = r[0]
        else:
            r1 = r[1]

        return self._react_interface(
            r0.composition, r1.composition, pd, grand_pd, calculators=calculators
        )

    def _get_rxn_iterable(self, combos, open_entries):
        return product(combos, [None])

    def _react_interface(self, r1, r2, pd, grand_pd=None, calculators=None):
        """Simple API for InterfacialReactivity module from pymatgen."""
        chempots = None

        if grand_pd:
            interface = InterfacialReactivity(
                r1,
                r2,
                grand_pd,
                norm=True,
                include_no_mixing_energy=False,
                pd_non_grand=pd,
                use_hull_energy=True,
            )
            chempots = grand_pd.chempots

        else:
            interface = InterfacialReactivity(
                r1,
                r2,
                pd,
                norm=False,
                include_no_mixing_energy=False,
                pd_non_grand=None,
                use_hull_energy=True,
            )

        rxns = []
        for _, _, _, rxn, _ in interface.get_kinks():
            if grand_pd:
                rxn = get_open_computed_rxn(rxn, pd.all_entries, chempots)
            else:
                rxn = get_computed_rxn(rxn, pd.all_entries)

            rxn = apply_calculators(rxn, calculators)
            rxns.append(rxn)

        return rxns


class MinimizeGrandPotentialEnumerator(MinimizeGibbsEnumerator):
    """
    Enumerator for finding all reactions between two reactants and an open element
    that are predicted by thermo; i.e., they appear when taking the
    convex hull along a straight line connecting any two phases in Phi-x
    phase space. Identity reactions are excluded.
    """

    def __init__(
        self,
        open_elem: Element,
        mu: float,
        precursors: Optional[List[str]] = None,
        target: Optional[str] = None,
        calculators: Optional[List[str]] = None,
    ):
        super().__init__(precursors=precursors, target=target, calculators=calculators)
        self.open_elem = Element(open_elem)  # type: ignore
        self.mu = mu
        self.chempots = {self.open_elem: self.mu}
        self._build_grand_pd = True

    def _react(self, reactants, products, calculators, pd=None, grand_pd=None):
        r = list(reactants)
        r0 = r[0]

        if len(r) == 1:
            r1 = r[0]
        else:
            r1 = r[1]

        open_elem = list(grand_pd.chempots.keys())[0]

        for reactant in r:
            elems = reactant.composition.elements
            if len(elems) == 1 and elems[0] == open_elem:  # skip if reactant = open_e
                return []

        return self._react_interface(
            r0.composition,
            r1.composition,
            pd,
            grand_pd=grand_pd,
            calculators=calculators,
        )



"""
Firetasks for acquiring ComputedEntry data from MPRester or another materials MongoDB
"""
import itertools
import warnings
from typing import Union, Optional, List

from fireworks import FiretaskBase, FWAction, explicit_serialize
from maggma.stores import MongoStore
from pymatgen.core.structure import Structure
from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry
from pymatgen.ext.matproj import MPRester
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.firetasks.utils import env_chk, get_logger

logger = get_logger(__name__)


@explicit_serialize
class EntriesFromMPRester(FiretaskBase):
    """
    Acquire ComputedStructureEntry objects from the Materials Project database.
    Automatically builds GibbsComputedEntry objects at specified temperature and
    filters by e_above_hull stability.

    Note: the PMG_MAPI_KEY enviornment variable must be properly configured on the
    computing resource.

    Required params:
        chemsys (str):
        temperature (float):
        e_above_hull (float):

    Optional params:
        include_polymorphs (bool):

    """

    required_params = ["chemsys", "temperature", "e_above_hull"]
    optional_params = ["include_polymorphs"]

    def run_task(self, fw_spec):
        chemsys = self["chemsys"]
        temperature = self["temperature"]
        e_above_hull = self["e_above_hull"]
        include_polymorphs = self.get("include_polymorphs", False)

        with MPRester() as mpr:
            entries = mpr.get_entries_in_chemsys(
                elements=chemsys, inc_structure="fiinal"
            )

        entries = process_entries(
            entries, temperature, e_above_hull, include_polymorphs
        )
        return FWAction(update_spec={"entries": entries})


@explicit_serialize
class EntriesFromDb(FiretaskBase):
    """
    Acquire ComputedStructureEntry objects from a custom materials MongoDB.
    Automatically builds GibbsComputedEntry objects at specified temperature and
    filters by e_above_hull stability.

    Required params:
        entry_db_file (str):
        chemsys (str):
        temperature (float):
        e_above_hull (float):

    Optional params:
        include_polymorphs (bool):

    """

    required_params = [
        "entry_db_file",
        "chemsys",
        "temperature",
        "e_above_hull",
    ]
    optional_params = [
        "include_polymorphs",
        "inc_structure",
        "compatible_only",
        "property_data",
    ]

    def run_task(self, fw_spec):
        db_file = env_chk(self["entry_db_file"], fw_spec)
        chemsys = self["chemsys"]
        temperature = self["temperature"]
        e_above_hull = self["e_above_hull"]
        include_polymorphs = self.get("include_polymorphs", False)
        inc_structure = self.get("inc_structure", "final")
        compatible_only = self.get("compatible_only", True)
        property_data = self.get("property_data", None)

        with MongoStore.from_db_file(db_file) as db:
            entries = get_all_entries_in_chemsys(
                db,
                chemsys,
                inc_structure=inc_structure,
                compatible_only=compatible_only,
                property_data=property_data,
                use_premade_entries=False,
            )

        entries = process_entries(
            entries, temperature, e_above_hull, include_polymorphs
        )
        return FWAction(update_spec={"entries": entries})


def process_entries(entries, temperature, e_above_hull, include_polymorphs):
    """

    Args:
        entries:
        temperature:
        e_above_hull:
        include_polymorphs:

    Returns:

    """
    entry_set = GibbsEntrySet.from_entries(entries=entries, temperature=temperature)
    entry_set = entry_set.filter_by_stability(
        e_above_hull=e_above_hull, include_polymorphs=include_polymorphs
    )
    return entry_set


def get_entries(  # noqa: C901
    db: MongoStore,
    chemsys_formula_id_criteria: Union[str, dict],
    compatible_only: bool = True,
    inc_structure: bool = None,
    property_data: Optional[List[str]] = None,
    use_premade_entries: bool = False,
    conventional_unit_cell: bool = False,
    sort_by_e_above_hull: bool = False,
):
    """
    Get a list of ComputedEntries or ComputedStructureEntries corresponding
    to a chemical system, formula, or materials_id or full criteria. Code adapted
    from pymatgen.ext.matproj.

    Args:
        db: MongoStore object with database connection
        chemsys_formula_id_criteria: A chemical system
            (e.g., Li-Fe-O), or formula (e.g., Fe2O3) or materials_id
            (e.g., mp-1234) or full Mongo-style dict criteria.
        compatible_only: Whether to return only "compatible"
            entries. Compatible entries are entries that have been
            processed using the MaterialsProjectCompatibility class,
            which performs adjustments to allow mixing of GGA and GGA+U
            calculations for more accurate phase diagrams and reaction
            energies.
        inc_structure: If None, entries returned are
            ComputedEntries. If inc_structure="initial",
            ComputedStructureEntries with initial structures are returned.
            Otherwise, ComputedStructureEntries with final structures
            are returned.
        property_data: Specify additional properties to include in
            entry.data. If None, no data. Should be a subset of
            supported_properties.
        use_premade_entries: Whether to use entry objects that have already been
            constructed. Defaults to False.
        conventional_unit_cell: Whether to get the standard
            conventional unit cell
        sort_by_e_above_hull: Whether to sort the list of entries by
            e_above_hull (will query e_above_hull as a property_data if True).
    Returns:
        List of ComputedEntry or ComputedStructureEntry objects.
    """
    params = [
        "deprecated",
        "run_type",
        "is_hubbard",
        "pseudo_potential",
        "hubbards",
        "potcar_symbols",
        "oxide_type",
    ]
    props = ["final_energy", "unit_cell_formula", "task_id"] + params
    if sort_by_e_above_hull:
        if property_data and "e_above_hull" not in property_data:
            property_data.append("e_above_hull")
        elif not property_data:
            property_data = ["e_above_hull"]
    if property_data:
        props += property_data
    if inc_structure:
        if inc_structure == "initial":
            props.append("initial_structure")
        else:
            props.append("structure")

    if not isinstance(chemsys_formula_id_criteria, dict):
        criteria = MPRester.parse_criteria(chemsys_formula_id_criteria)
    else:
        criteria = chemsys_formula_id_criteria

    if use_premade_entries:
        props = ["entries", "deprecated"]

    entries = []
    for d in db.query(criteria, props):
        if d.get("deprecated"):
            continue
        if use_premade_entries:
            ent = d["entries"]
            if ent.get("GGA"):
                e = ComputedStructureEntry.from_dict(ent["GGA"])
            elif ent.get("GGA+U"):
                e = ComputedStructureEntry.from_dict(ent["GGA+U"])
            else:
                print(f"Missing entry for {d['_id']}")
                continue
        else:
            d["potcar_symbols"] = [
                "%s %s" % (d["pseudo_potential"]["functional"], l)
                for l in d["pseudo_potential"].get("labels", [])
            ]
            data = {"oxide_type": d["oxide_type"]}
            if property_data:
                data.update({k: d[k] for k in property_data})
            if not inc_structure:
                e = ComputedEntry(
                    d["unit_cell_formula"],
                    d["final_energy"],
                    parameters={k: d[k] for k in params},
                    data=data,
                    entry_id=d["task_id"],
                )
            else:
                prim = Structure.from_dict(
                    d["initial_structure"]
                    if inc_structure == "initial"
                    else d["structure"]
                )
                if conventional_unit_cell:
                    s = SpacegroupAnalyzer(prim).get_conventional_standard_structure()
                    energy = d["final_energy"] * (len(s) / len(prim))
                else:
                    s = prim.copy()
                    energy = d["final_energy"]
                e = ComputedStructureEntry(
                    s,
                    energy,
                    parameters={k: d[k] for k in params},
                    data=data,
                    entry_id=d["task_id"],
                )
        entries.append(e)
    if compatible_only:
        from pymatgen.entries.compatibility import MaterialsProject2020Compatibility

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="Failed to guess oxidation states.*"
            )
            entries = MaterialsProject2020Compatibility().process_entries(
                entries, clean=True
            )
    if sort_by_e_above_hull:
        entries = sorted(entries, key=lambda entry: entry.data["e_above_hull"])
    return entries


def get_all_entries_in_chemsys(
    db,
    elements,
    compatible_only=True,
    inc_structure=None,
    property_data=None,
    use_premade_entries=False,
    conventional_unit_cell=False,
    n=1000,
):
    """
    Helper method for getting all entries in a total chemical system by querying
    database for all sub-chemical systems. Code adadpted from pymatgen.ext.matproj
    and modified to support very large chemical systems.

    Args:
        elements (str or [str]): Chemical system string comprising element
            symbols separated by dashes, e.g., "Li-Fe-O" or List of element
            symbols, e.g., ["Li", "Fe", "O"].
        compatible_only (bool): Whether to return only "compatible"
            entries. Compatible entries are entries that have been
            processed using the MaterialsProjectCompatibility class,
            which performs adjustments to allow mixing of GGA and GGA+U
            calculations for more accurate phase diagrams and reaction
            energies.
        inc_structure (str): If None, entries returned are
            ComputedEntries. If inc_structure="final",
            ComputedStructureEntries with final structures are returned.
            Otherwise, ComputedStructureEntries with initial structures
            are returned.
        property_data (list): Specify additional properties to include in
            entry.data. If None, no data. Should be a subset of
            supported_properties.
        conventional_unit_cell (bool): Whether to get the standard
            conventional unit cell
        n (int): Chunk size, i.e., number of sub-chemical systems to consider
    Returns:
        List of ComputedEntries.
    """

    def divide_chunks(l, n):
        for i in range(0, len(l), n):
            yield l[i : i + n]

    if isinstance(elements, str):
        elements = elements.split("-")

    if len(elements) < 13:
        all_chemsyses = []
        for i in range(len(elements)):
            for els in itertools.combinations(elements, i + 1):
                all_chemsyses.append("-".join(sorted(els)))

        all_chemsyses = list(divide_chunks(all_chemsyses, n))

        entries = []
        for chemsys_group in all_chemsyses:
            entries.extend(
                get_entries(
                    db,
                    {"chemsys": {"$in": chemsys_group}},
                    compatible_only=compatible_only,
                    inc_structure=inc_structure,
                    property_data=property_data,
                    use_premade_entries=use_premade_entries,
                    conventional_unit_cell=conventional_unit_cell,
                )
            )
    else:
        entries = get_entries(
            db,
            {"elements": {"$not": {"$elemMatch": {"$nin": elements}}}},
            compatible_only=compatible_only,
            inc_structure=inc_structure,
            property_data=property_data,
            use_premade_entries=use_premade_entries,
            conventional_unit_cell=conventional_unit_cell,
        )

    return entries



"""
Utility functions used in the reaction classes.
"""

import numpy as np
from pymatgen.core.composition import Element


def is_separable_rxn(rxn, target_comp, added_elems):
    """

    Args:
        rxn:
        target_comp:
        added_elems:

    Returns:

    """
    added_elems = [Element(e) for e in added_elems.split("-")]
    products = rxn.products.copy()
    products.remove(target_comp)
    separable = np.array(
        [set(comp.elements).issubset(added_elems) for comp in products]
    )
    found = False
    if separable.all():
        found = True
    return found



"""
Entry and entry set classes used in reaction enumeration / networks
"""
from rxn_network.entries.gibbs import GibbsComputedEntry
from rxn_network.entries.nist import NISTReferenceEntry
from rxn_network.entries.entry_set import GibbsEntrySet



"""
Basic interface for a (chemical) Reaction
"""
from abc import ABCMeta, abstractmethod
from typing import List

import numpy as np
from monty.json import MSONable
from pymatgen.core.composition import Composition, Element


class Reaction(MSONable, metaclass=ABCMeta):
    """
    Base definition for a Reaction
    """

    @property
    @abstractmethod
    def reactants(self) -> List[Composition]:
        """List of reactants for this reaction"""

    @property
    @abstractmethod
    def products(self) -> List[Composition]:
        """List of products for this reaction"""

    @property
    @abstractmethod
    def coefficients(self) -> np.ndarray:
        """Coefficients of the reaction"""

    @property
    @abstractmethod
    def energy(self) -> float:
        """The energy of this reaction in total eV"""

    @property
    def compositions(self) -> List[Composition]:
        """List of all compositions in the reaction"""
        return self.reactants + self.products

    @property
    def elements(self) -> List[Element]:
        """List of elements in the reaction"""
        return list(set(el for comp in self.compositions for el in comp.elements))

    @property
    def num_atoms(self) -> float:
        """Total number of atoms in this reaction"""
        return (
            sum(
                [
                    comp[element] * abs(coeff)
                    for element in self.elements
                    for coeff, comp in zip(self.coefficients, self.compositions)
                ]
            )
            / 2
        )

    @property
    def energy_per_atom(self) -> float:
        """The energy per atom of this reaction in eV"""
        return self.energy / self.num_atoms



"""
A calculator class for determining chemical potential distance of reactions
"""
from itertools import chain, combinations, product
from typing import List, Optional

import numpy as np
from pymatgen.analysis.phase_diagram import PDEntry

from rxn_network.core.calculator import Calculator
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.thermo.chempot_diagram import ChemicalPotentialDiagram


class ChempotDistanceCalculator(Calculator):
    """
    Calculator for determining the "chemical potential distance" for a reaction
    (in eV/atom).

    For more information on this specific implementation of the algorithm,
    please cite/reference the paper below:

    Todd, Paul K., McDermott, M.J., et al. “Selectivity in yttrium manganese oxide
    synthesis via local chemical potentials in hyperdimensional phase space.”
    ArXiv:2104.05986 [Cond-Mat], Apr. 2021. arXiv.org, http://arxiv.org/abs/2104.05986
    """

    def __init__(
        self,
        cpd: ChemicalPotentialDiagram,
        mu_func: str = "sum",
        name: str = "chempot_distance",
    ):
        """
        Args:
            cpd: the chemical potential diagram
            mu_func: the name of the function used to process the interfacial
                chemical potential distances into a single value describing the whole
                reaction.
            name: the data dictionary key with which to store the calculated value.
        """
        self.cpd = cpd
        self.name = name

        if mu_func == "max":
            self._mu_func = np.max  # type: ignore
        elif mu_func == "mean":
            self._mu_func = np.mean  # type: ignore
        elif mu_func == "sum":
            self._mu_func = np.sum  # type: ignore

    def calculate(self, rxn: ComputedReaction) -> float:
        """
        Calculates the chemical potential distance in eV/atom. The mu_func parameter
        determines how the distance is calculated for the overall reaction.

        Args:
            rxn: the reaction object

        Returns:
            The chemical potential distance of the reaction.
        """
        combos = chain(
            product(rxn.reactant_entries, rxn.product_entries),
            combinations(rxn.product_entries, 2),
        )
        distances = [
            self.cpd.shortest_domain_distance(
                combo[0].composition.reduced_formula,
                combo[1].composition.reduced_formula,
            )
            for combo in combos
        ]

        distance = float(self._mu_func(distances))
        return distance

    def decorate(self, rxn: ComputedReaction) -> ComputedReaction:
        """
        Decorates the reaction (in place) with the chemical potential distance by
        storing the value within the reaction's data dictionary.

        Args:
            rxn: The reaction object.

        Returns:
            The reaction object, modified in place
        """
        if not rxn.data:
            rxn.data = {}

        rxn.data[self.name] = self.calculate(rxn)

        return rxn

    @classmethod
    def from_entries(
        cls,
        entries: List[PDEntry],
        mu_func: str = "sum",
        name: str = "chempot_distance",
        **kwargs
    ) -> "ChempotDistanceCalculator":
        """
        Convenience constructor which first builds the ChemicalPotentialDiagram
        object from a list of entry objects.

        Args:
            entries: entry objects used to build the ChemicalPotentialDiagram
            mu_func: the name of the function used to process the interfacial
                chemical potential distances into a single value describing the whole
                reaction.
            name: the data dictionary key by which to store the calculated value,
                defaults to "chempot_distance"
            **kwargs: optional kwargs passed to ChemicalPotentialDiagram; defaults to
                "default_min_limit"=-50

        Returns:
            A ChempotDistanceCalculator object
        """
        if not kwargs.get("default_min_limit"):
            kwargs["default_min_limit"] = -50

        cpd = ChemicalPotentialDiagram(entries=entries, **kwargs)
        return cls(cpd, mu_func, name)

    @property
    def mu_func(self):
        """Returns the function used to process the interfacial mu distances"""
        return self._mu_func



""" Tests for Softplus """
import pytest
from rxn_network.costs.softplus import Softplus


@pytest.fixture(scope="module")
def softplus_with_attr():
    return Softplus(temp=800, params=["energy_per_atom"], weights=[1.0])


@pytest.fixture(scope="module")
def softplus_with_attr_and_param():
    return Softplus(
        temp=800, params=["energy_per_atom", "test_param"], weights=[0.3, 0.7]
    )


def test_evaluate(softplus_with_attr, softplus_with_attr_and_param, computed_rxn):
    r = computed_rxn.copy()
    r.data = {"test_param": 0.1}

    cost1 = softplus_with_attr.evaluate(r)
    cost2 = softplus_with_attr_and_param.evaluate(r)

    assert cost1 == pytest.approx(0.250771199)
    assert cost2 == pytest.approx(0.297691790)


def test_missing_parameter(softplus_with_attr_and_param, computed_rxn):
    with pytest.raises(ValueError) as error:
        assert softplus_with_attr_and_param.evaluate(computed_rxn)
    assert str(error.value) == "Reaction is missing parameter test_param!"



"""
Implementation of Fireworks for performing reaction enumreation and network
construction
"""
from fireworks import Firework

from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.firetasks.build_inputs import EntriesFromDb, EntriesFromMPRester
from rxn_network.firetasks.parse_outputs import ReactionsToDb

from rxn_network.firetasks.run_calc import RunEnumerators


class EnumeratorFW(Firework):
    """
    Firework for running a list of enumerators (which outputs a list of reactions).
    """

    def __init__(
        self,
        enumerators,
        entries=None,
        chemsys=None,
        temperature=None,
        e_above_hull=None,
        db_file=None,
        entry_db_file=None,
        include_polymorphs=False,
        parents=None,
    ):
        """

        Args:
            enumerators:
            entries:
            chemsys:
            temperature:
            e_above_hull:
            db_file:
            entry_db_file:
            include_polymorphs:
            parents:
        """

        tasks = []

        entry_set = None
        if entries:
            entry_set = GibbsEntrySet(entries)
            chemsys = "-".join(sorted(list(entry_set.chemsys)))
        else:
            if entry_db_file:
                entry_task = EntriesFromDb(
                    entry_db_file=entry_db_file,
                    chemsys=chemsys,
                    temperature=temperature,
                    e_above_hull=e_above_hull,
                    include_polymorphs=include_polymorphs,
                )
            else:
                entry_task = EntriesFromMPRester(
                    chemsys=chemsys,
                    temperature=temperature,
                    e_above_hull=e_above_hull,
                    include_polymorphs=include_polymorphs,
                )
            tasks.append(entry_task)

        targets = {enumerator.target for enumerator in enumerators}
        if len(targets) != 1:
            raise ValueError("Enumerators contain different targets!")
        target = targets.pop()
        fw_name = f"Reaction Enumeration (Target: {target}): {chemsys}"

        tasks.append(
            RunEnumerators(enumerators=enumerators, entries=entry_set, chemsys=chemsys)
        )
        tasks.append(ReactionsToDb(db_file=db_file, calc_dir="."))

        super().__init__(tasks, parents=parents, name=fw_name)


class NetworkFW(Firework):
    pass



"""
Implementations of reaction enumerators
"""
from rxn_network.enumerators.basic import BasicEnumerator, BasicOpenEnumerator
from rxn_network.enumerators.minimize import (
    MinimizeGibbsEnumerator,
    MinimizeGrandPotentialEnumerator,
)



#!/usr/bin/env python


from pathlib import Path

from setuptools import find_packages, setup

module_dir = Path(__file__).resolve().parent

with open(module_dir / "README.md") as f:
    long_desc = f.read()

setup(
    name="reaction-network",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="Reaction-network is a Python package for predicting likely inorganic "
    "chemical reaction pathways using graph theory.",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url="https://github.com/GENESIS-EFRC/reaction-network",
    author="Matthew McDermott",
    author_email="mcdermott@lbl.gov",
    maintainer="Matthew McDermott",
    license="modified BSD",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"rxn_network": ["py.typed"]},
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        "setuptools",
        "fireworks>=1.9.7",
        "maggma>=0.26.0",
        "numba>= 0.52.0",
        "pymatgen>=2022.0.10",
    ],
    extra_requires={"demo": ["jupyter>=1.0.0"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    tests_require=["pytest"],
    python_requires=">=3.8",
)



""" Implementations of various reaction classes for the reaction-network """
from rxn_network.reactions.basic import BasicReaction
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.open import OpenComputedReaction
from rxn_network.reactions.reaction_set import ReactionSet



"""
A reaction class that builds reactions based on ComputedEntry objects and provides
information about reaction thermodynamics.
"""
from typing import Dict, List, Optional, Union

import numpy as np
from pymatgen.core.composition import Composition
from pymatgen.entries.computed_entries import ComputedEntry
from uncertainties import ufloat

from rxn_network.reactions.basic import BasicReaction


class ComputedReaction(BasicReaction):
    """
    Convenience class to generate a reaction from ComputedEntry objects, with
    some additional attributes, such as a reaction energy based on computed
    energies. This class also balances the reaction.
    """

    def __init__(
        self,
        entries: List[ComputedEntry],
        coefficients: Union[np.ndarray, List[float]],
        data: Optional[Dict] = None,
        lowest_num_errors: Union[int, float] = 0,
    ):
        """
        Args:
            entries: List of ComputedEntry objects.
            coefficients: List of reaction coefficients.
            data: Optional dict of data
            lowest_num_errors: number of "errors" encountered during reaction balancing
        """
        self._entries = list(entries)
        self.reactant_entries = [
            entry for entry, coeff in zip(entries, coefficients) if coeff < 0
        ]
        self.product_entries = [
            entry for entry, coeff in zip(entries, coefficients) if coeff > 0
        ]
        compositions = [e.composition.reduced_composition for e in entries]

        super().__init__(
            compositions, coefficients, data=data, lowest_num_errors=lowest_num_errors
        )

    @classmethod
    def balance(
        cls,
        reactant_entries: List[ComputedEntry],
        product_entries: List[ComputedEntry],
        data: Optional[Dict] = None,
    ):  # pylint: disable = W0221
        """
        Balances and returns a new ComputedReaction.

        Reactants and products to be specified as list of
        pymatgen.core.structure.Composition.  e.g., [comp1, comp2]

        Args:
            reactant_entries: List of reactant entries
            product_entries: List of product entries
            data: Optional dict of data
        """
        reactant_comps = [e.composition.reduced_composition for e in reactant_entries]
        product_comps = [e.composition.reduced_composition for e in product_entries]
        coefficients, lowest_num_errors = cls._balance_coeffs(
            reactant_comps, product_comps
        )

        return cls(
            entries=list(reactant_entries) + list(product_entries),
            coefficients=coefficients,
            data=data,
            lowest_num_errors=lowest_num_errors,
        )

    @property
    def energy(self) -> float:
        """
        Returns (float):
            The calculated reaction energy.
        """
        calc_energies: Dict[Composition, float] = {}

        for entry in self._entries:
            (comp, factor) = entry.composition.get_reduced_composition_and_factor()
            calc_energies[comp] = min(
                calc_energies.get(comp, float("inf")), entry.energy / factor
            )

        return sum(
            [
                amt * calc_energies[c]
                for amt, c in zip(self.coefficients, self.compositions)
            ]
        )

    @property
    def energy_per_atom(self) -> float:
        """
        Returns (float):
            The calculated reaction energy in eV, divided by the total number of
            atoms in the reaction.
        """
        return self.energy / self.num_atoms

    @property
    def energy_uncertainty(self):
        """
        Calculates the uncertainty in the reaction energy based on the uncertainty in the
        energies of the reactants/products
        """

        calc_energies = {}

        for entry in self._entries:
            (comp, factor) = entry.composition.get_reduced_composition_and_factor()
            energy_ufloat = ufloat(entry.energy, entry.correction_uncertainty)
            calc_energies[comp] = min(
                calc_energies.get(comp, float("inf")), energy_ufloat / factor
            )

        energy_with_uncertainty = sum(
            [
                amt * calc_energies[c]
                for amt, c in zip(self.coefficients, self.compositions)
            ]
        )

        return energy_with_uncertainty.s

    @property
    def energy_uncertainty_per_atom(self):
        return self.energy_uncertainty / self.num_atoms

    @property
    def entries(self):
        """
        Returns a copy of the entries
        """
        return self._entries

    def copy(self) -> "ComputedReaction":
        """
        Returns a copy of the Reaction object
        """
        return ComputedReaction(
            self.entries, self.coefficients, self.data, self.lowest_num_errors
        )

    def reverse(self):
        """
        Returns a reversed reaction (i.e. sides flipped)

        """
        return ComputedReaction(
            self.entries, -1 * self.coefficients, self.data, self.lowest_num_errors
        )

    def __hash__(self):
        return BasicReaction.__hash__(self)

    def __eq__(self, other):
        eq = BasicReaction.__eq__(self, other)
        if not eq:
            return False
        else:
            return np.isclose(self.energy_per_atom, other.energy_per_atom)



"""
Utility Fireworks functions borrowed from the atomate package
"""
import logging
import os
import sys
from typing import Optional


def env_chk(
    val: str,
    fw_spec: dict,
    strict: Optional[bool] = True,
    default: Optional[str] = None,
):
    """
    Code borrowed from the atomate package.

    env_chk() is a way to set different values for a property depending
    on the worker machine. For example, you might have slightly different
    executable names or scratch directories on different machines.

    Args:
        val: any value, with ">><<" notation reserved for special env lookup values
        fw_spec: fw_spec where one can find the _fw_env keys
        strict: if True, errors if env format (>><<) specified but cannot be found in fw_spec
        default: if val is None or env cannot be found in non-strict mode,
                 return default
    """
    if val is None:
        return default

    if isinstance(val, str) and val.startswith(">>") and val.endswith("<<"):
        if strict:
            return fw_spec["_fw_env"][val[2:-2]]
        return fw_spec.get("_fw_env", {}).get(val[2:-2], default)
    return val


def get_logger(
    name: str,
    level=logging.DEBUG,
    log_format="%(asctime)s %(levelname)s %(name)s %(message)s",
    stream=sys.stdout,
):
    """
    Code borrowed from the atomate package.

    Helper method for acquiring logger.
    """

    logger = logging.getLogger(name)
    logger.setLevel(level)

    formatter = logging.Formatter(log_format)

    sh = logging.StreamHandler(stream=stream)
    sh.setFormatter(formatter)

    logger.addHandler(sh)

    return logger



""" Core interfaces for the reaction-network package. """
from rxn_network.core.calculator import Calculator
from rxn_network.core.cost_function import CostFunction
from rxn_network.core.enumerator import Enumerator
from rxn_network.core.network import Network
from rxn_network.core.pathway import Pathway
from rxn_network.core.reaction import Reaction
from rxn_network.core.solver import Solver



""" Tests for GibbsEntrySet. """
import pytest
from pymatgen.analysis.phase_diagram import PhaseDiagram

from rxn_network.entries.entry_set import GibbsEntrySet


@pytest.mark.parametrize(
    "e_above_hull, expected_phases",
    [
        (
            0.030,
            {
                "Mn",
                "Mn2O3",
                "Mn3O4",
                "Mn5O8",
                "MnO",
                "MnO2",
                "O2",
                "Y",
                "Y2Mn2O7",
                "Y2O3",
                "YMn12",
                "YMn2O5",
                "YMnO3",
            },
        ),
        (
            0.100,
            {
                "YMnO3",
                "Y",
                "Y2O3",
                "Mn5O8",
                "Mn",
                "Y2Mn2O7",
                "YMn2O4",
                "YMn2O5",
                "MnO2",
                "Mn21O40",
                "Mn3O4",
                "Mn7O12",
                "MnO",
                "O2",
                "Mn2O3",
                "YMn12",
            },
        ),
    ],
)
def test_filter_by_stability(e_above_hull, expected_phases, gibbs_entries):
    filtered_entries = gibbs_entries.filter_by_stability(e_above_hull)

    actual_phases = {e.composition.reduced_formula for e in filtered_entries}

    assert actual_phases == expected_phases


def test_build_indices(gibbs_entries):
    entries = gibbs_entries.copy()
    entries.build_indices()

    num_entries = len(entries)

    indices = [e.data["idx"] for e in entries.entries_list]
    assert indices == list(range(num_entries))


def test_get_min_entry_by_formula(gibbs_entries):
    f_id = [("YMnO3", "mp-19385"), ("Mn2O3", "mp-1172875"), ("MnO2", "mp-1279979")]
    for f, entry_id in f_id:
        assert gibbs_entries.get_min_entry_by_formula(f).entry_id == entry_id


def test_stabilize_entry(gibbs_entries):
    entries = gibbs_entries.copy()

    formulas = ["YMn3O6", "Mn2O5"]

    for f in formulas:
        e = entries.get_min_entry_by_formula(f)
        e_stable = entries.stabilize_entry(e)
        entries.add(e_stable)

        assert e_stable in PhaseDiagram(entries).stable_entries


def test_from_pd(mp_entries):
    entries = GibbsEntrySet.from_pd(PhaseDiagram(mp_entries), temperature=1000)
    assert entries is not None



"""
Basic interface for a reaction Enumerator
"""
import logging
from abc import ABCMeta, abstractmethod
from typing import List

from monty.json import MSONable


class Enumerator(MSONable, metaclass=ABCMeta):
    """
    Base definition for the reaction enumeration methodology
    """

    def __init__(self, precursors, target, calculators):
        self.logger = logging.getLogger(str(self.__class__.__name__))
        self.logger.setLevel("INFO")
        self.precursors = precursors
        self.target = target

        if not calculators:
            calculators = []

        self.calculators = calculators

    @abstractmethod
    def enumerate(self, entries):
        """
        Enumerates the potential reactions from the list of entries
        """

    @abstractmethod
    def estimate_max_num_reactions(self, entries) -> int:
        """
        Estimate of the number of reactions from a list of entries
        """



"""
An entry set class for acquiring entries with Gibbs formation energies
"""
from typing import List, Optional, Union, Set, Dict

from monty.json import MontyDecoder
from pymatgen.analysis.phase_diagram import PhaseDiagram
from pymatgen.core import Composition
from pymatgen.entries.computed_entries import (
    ComputedEntry,
    ComputedStructureEntry,
    ConstantEnergyAdjustment,
)
from pymatgen.entries.entry_tools import EntrySet
from tqdm.auto import tqdm

from rxn_network.entries.gibbs import GibbsComputedEntry
from rxn_network.entries.nist import NISTReferenceEntry
from rxn_network.thermo.utils import expand_pd


class GibbsEntrySet(EntrySet):
    """
    An extension of pymatgen's EntrySet to include factory methods for constructing
    GibbsComputedEntry objects from zero-temperature ComputedStructureEntry objects.
    """

    def __init__(self, entries: List[Union[GibbsComputedEntry, NISTReferenceEntry]]):
        """
        The supplied collection of entries will automatically be converted to a set of
        unique entries.

        Args:
            entries: A collection of entry objects that will make up the entry set.
        """
        super().__init__(entries)
        self.entries_list = list(
            sorted(entries, key=lambda e: e.composition.reduced_formula)
        )
        self.build_indices()

    def filter_by_stability(
        self, e_above_hull: float, include_polymorphs: Optional[bool] = False
    ) -> "GibbsEntrySet":
        """
        Filter the entry set by a metastability (energy above hull) cutoff.

        Args:
            e_above_hull: Energy above hull, the cutoff describing the allowed
                metastability of the entries as determined via phase diagram
                construction.
            include_polymorphs: optional specification of whether to include
                metastable polymorphs. Defaults to False.

        Returns:
            A new GibbsEntrySet where the entries have been filtered by an energy
            cutoff (e_above_hull) via phase diagram construction.
        """
        pd_dict = expand_pd(self.entries)

        filtered_entries: Set[Union[GibbsComputedEntry, NISTReferenceEntry]] = set()
        all_comps: Dict[str, Union[GibbsComputedEntry, NISTReferenceEntry]] = dict()

        for chemsys, pd in pd_dict.items():
            for entry in pd.all_entries:
                if (
                    entry in filtered_entries
                    or pd.get_e_above_hull(entry) > e_above_hull
                ):
                    continue

                formula = entry.composition.reduced_formula
                if not include_polymorphs and (formula in all_comps):
                    if all_comps[formula].energy_per_atom < entry.energy_per_atom:
                        continue
                    filtered_entries.remove(all_comps[formula])

                all_comps[formula] = entry
                filtered_entries.add(entry)

        return self.__class__(list(filtered_entries))

    def build_indices(self):
        for idx, e in enumerate(self.entries_list):
            e.data.update({"idx": idx})

    def get_min_entry_by_formula(self, formula: str) -> ComputedEntry:
        """
        Helper method for acquiring the ground state entry with the specified formula.

        Args:
            formula: The chemical formula of the desired entry.

        Returns:
            Ground state computed entry object.
        """
        comp = Composition(formula).reduced_composition
        possible_entries = filter(
            lambda x: x.composition.reduced_composition == comp, self.entries
        )
        return sorted(possible_entries, key=lambda x: x.energy_per_atom)[0]

    def stabilize_entry(self, entry: ComputedEntry, tol: float = 1e-6) -> ComputedEntry:
        """
        Helper method for lowering the energy of a single entry such that it is just
        barely stable on the phase diagram.

        Args:
            entry: A computed entry object.
            tol: The numerical padding added to the energy correction to guarantee
                that it is determined to be stable during phase diagram construction.

        Returns:
            A new ComputedEntry with energy adjustment making it appear to be stable.
        """
        chemsys = [str(e) for e in entry.composition.elements]
        entries = self.get_subset_in_chemsys(chemsys)
        pd = PhaseDiagram(entries)
        e_above_hull = pd.get_e_above_hull(entry)

        if e_above_hull == 0.0:
            new_entry = entry
        else:
            e_adj = -1 * pd.get_e_above_hull(entry) * entry.composition.num_atoms - tol
            adjustment = ConstantEnergyAdjustment(
                value=e_adj,
                name="Stabilization Adjustment",
                description="Shifts energy so that " "entry is on the convex hull",
            )

            entry_dict = entry.as_dict()
            entry_dict["energy_adjustments"].append(adjustment)
            new_entry = MontyDecoder().process_decoded(entry_dict)

        return new_entry

    @classmethod
    def from_pd(cls, pd: PhaseDiagram, temperature: float) -> "GibbsEntrySet":
        """
        Constructor method for building a GibbsEntrySet from an existing phase diagram.

        Args:
            pd: Phase Diagram object (pymatgen)
            temperature: Temperature [K] for determining Gibbs Free Energy of
                formation, dGf(T)

        Returns:
            A GibbsEntrySet containing a collection of GibbsComputedEntry and
            experimental reference entry objects at the specified temperature.

        """
        gibbs_entries = []
        for entry in pd.all_entries:
            if entry.composition.is_element and entry not in pd.el_refs.values():
                continue
            composition = entry.composition

            if composition.reduced_formula in NISTReferenceEntry.REFERENCES:
                new_entry = NISTReferenceEntry(
                    composition=composition, temperature=temperature
                )
            else:
                structure = entry.structure
                formation_energy_per_atom = pd.get_form_energy_per_atom(entry)

                new_entry = GibbsComputedEntry.from_structure(
                    structure=structure,
                    formation_energy_per_atom=formation_energy_per_atom,
                    temperature=temperature,
                    energy_adjustments=None,
                    parameters=entry.parameters,
                    data=entry.data,
                    entry_id=entry.entry_id,
                )

            gibbs_entries.append(new_entry)

        return cls(gibbs_entries)

    @classmethod
    def from_entries(
        cls, entries: List[ComputedStructureEntry], temperature: float
    ) -> "GibbsEntrySet":
        """
        Constructor method for initializing GibbsEntrySet from T = 0 K
        ComputedStructureEntry objects, as acquired from a thermochemical
        database e.g. The Materials Project. Automatically expands the phase
        diagram for large chemical systems (10 or more elements) to avoid limitations
        of Qhull.

        Args:
            entries: List of ComputedStructureEntry objects, as downloaded from The
                Materials Project API.
            temperature: Temperature for estimating Gibbs free energy of formation [K]

        Returns:
            A GibbsEntrySet containing a collection of GibbsComputedEntry and
            experimental reference entry objects at the specified temperature.
        """
        e_set = EntrySet(entries)
        new_entries: Set[GibbsComputedEntry] = set()
        if len(e_set.chemsys) <= 9:  # Qhull algorithm struggles beyond 9 dimensions
            pd = PhaseDiagram(e_set)
            return cls.from_pd(pd, temperature)

        pd_dict = expand_pd(list(e_set))
        for chemsys, pd in tqdm(pd_dict.items()):
            gibbs_set = cls.from_pd(pd, temperature)
            new_entries.update(gibbs_set)

        return cls(list(new_entries))

    def copy(self) -> "GibbsEntrySet":
        return GibbsEntrySet(entries=self.entries)



""" Tests for BasicEnumerator and BasicOpenEnumerator """
from pathlib import Path
from monty.serialization import loadfn
import pytest

from rxn_network.enumerators.basic import BasicEnumerator, BasicOpenEnumerator

TEST_FILES_PATH = Path(__file__).parent.parent / "test_files"
RXNS_FILE = "ymno3_rxns.json.gz"


@pytest.fixture
def ymno3_rxns():
    return loadfn(TEST_FILES_PATH / RXNS_FILE)


@pytest.fixture
def basic_enumerator_default():
    return BasicEnumerator()


@pytest.fixture
def basic_enumerator_with_calculator():
    return BasicEnumerator(calculators=["ChempotDistanceCalculator"])


@pytest.fixture
def basic_enumerator_with_precursors():
    return BasicEnumerator(precursors=["Y2O3", "Mn2O3"])


@pytest.fixture
def basic_enumerator_with_target():
    return BasicEnumerator(target="YMnO3")


@pytest.fixture
def basic_enumerator_with_precursors_and_target():
    return BasicEnumerator(precursors=["Y2O3", "Mn2O3"], target="YMnO3")


@pytest.fixture
def basic_open_enumerator():
    return BasicOpenEnumerator(["O2"])


@pytest.fixture
def basic_open_enumerator_with_precursors():
    return BasicOpenEnumerator(["O2"], precursors=["Y2O3", "Mn2O3"])


@pytest.fixture
def basic_open_enumerator_with_target():
    return BasicOpenEnumerator(["O2"], target="Y2Mn2O7")


@pytest.fixture
def basic_open_enumerator_with_precursors_and_target():
    return BasicOpenEnumerator(["O2"], precursors=["Y2O3", "Mn2O3"], target="Y2Mn2O7")


def test_enumerate(
    filtered_entries, basic_enumerator_default, basic_enumerator_with_calculator
):
    expected_num_rxns = 538

    for enumerator in [basic_enumerator_default, basic_enumerator_with_calculator]:
        rxns = enumerator.enumerate(filtered_entries)

        assert expected_num_rxns == len(rxns)
        assert len(rxns) == len(set(rxns))
        assert all([not r.is_identity for r in rxns])

        if enumerator.calculators:
            assert all([r.data["chempot_distance"] is not None for r in rxns])


def test_enumerate_with_precursors(
    filtered_entries,
    basic_enumerator_with_precursors,
    basic_open_enumerator_with_precursors,
):

    for enumerator in [
        basic_enumerator_with_precursors,
        basic_open_enumerator_with_precursors,
    ]:

        rxns = enumerator.enumerate(filtered_entries)
        precursors = enumerator.precursors

        for r in rxns:
            reactants = [i.reduced_formula for i in r.reactants]
            products = [i.reduced_formula for i in r.products]

            for precursor in precursors:
                assert precursor in reactants
                assert precursor not in products


def test_enumerate_with_target(
    filtered_entries, basic_enumerator_with_target, basic_open_enumerator_with_target
):

    for enumerator in [basic_enumerator_with_target, basic_open_enumerator_with_target]:
        rxns = enumerator.enumerate(filtered_entries)
        target = enumerator.target

        for r in rxns:
            reactants = [i.reduced_formula for i in r.reactants]
            products = [i.reduced_formula for i in r.products]
            assert target not in reactants
            assert target in products


def test_enumerate_with_precursors_and_target(
    filtered_entries, basic_enumerator_with_precursors_and_target
):
    rxns = basic_enumerator_with_precursors_and_target.enumerate(filtered_entries)

    assert len(rxns) == 1
    assert str(rxns[0]) == "Mn2O3 + Y2O3 -> 2 YMnO3"


def test_open_enumerate_with_precursors_and_target(
    filtered_entries, basic_open_enumerator_with_precursors_and_target
):
    rxns = basic_open_enumerator_with_precursors_and_target.enumerate(filtered_entries)

    assert len(rxns) == 11

    for r in rxns:
        reactants = [i.reduced_formula for i in r.reactants]
        products = [i.reduced_formula for i in r.products]
        assert "O2" in reactants or "O2" in products
        assert "Y2Mn2O7" in products


def test_open_enumerate(filtered_entries, basic_open_enumerator):
    expected_num_rxns = 1358

    rxns = basic_open_enumerator.enumerate(filtered_entries)
    print(rxns)

    assert expected_num_rxns == len(rxns)
    assert len(rxns) == len(set(rxns))
    assert all([not r.is_identity for r in rxns])



""" Tests for OpenComputedReaction. """
import pytest

from pymatgen.core.composition import Element
from rxn_network.reactions.open import OpenComputedReaction


@pytest.fixture(scope="module", params=["Na", "Mn", "O"])
def element(request):
    return request.param


@pytest.fixture(scope="module", params=[0, -2, -4, -6])
def chempot(request):
    return request.param


@pytest.fixture(scope="module")
def open_computed_rxn(computed_rxn, element, chempot):
    chempots = {Element(element): chempot}

    return OpenComputedReaction(
        entries=computed_rxn.entries,
        coefficients=computed_rxn.coefficients,
        chempots=chempots,
        data=computed_rxn.data,
        lowest_num_errors=computed_rxn.lowest_num_errors,
    )


def test_energy(open_computed_rxn):
    open_elem = list(open_computed_rxn.chempots.keys())[0]
    chempot = list(open_computed_rxn.chempots.values())[0]

    expected = {
        "Na": {
            0: -2.701048424999957,
            -2: -2.701048424999957,
            -4: -2.701048424999957,
            -6: -2.701048424999957,
        },
        "Mn": {
            0: -2.701048424999957,
            -2: -2.701048424999957,
            -4: -2.701048424999957,
            -6: -2.701048424999957,
        },
        "O": {
            0: -2.701048424999957,
            -2: -0.7010484249999713,
            -4: 1.2989515750000287,
            -6: 3.2989515750000145,
        },
    }

    assert open_computed_rxn.energy == pytest.approx(expected[open_elem.name][chempot])


def test_num_atoms(open_computed_rxn):
    open_elem = list(open_computed_rxn.chempots.keys())[0]

    expected = {"Na": 13, "Mn": 13, "O": 8}
    assert open_computed_rxn.num_atoms == pytest.approx(expected[open_elem.name])


def test_elements(open_computed_rxn):
    open_elem = list(open_computed_rxn.chempots.keys())[0]

    expected = {
        "Na": {Element(e) for e in ["Mn", "O", "Y", "Cl"]},
        "Mn": {Element(e) for e in ["Na", "O", "Y", "Cl"]},
        "O": {Element(e) for e in ["Na", "Mn", "Y", "Cl"]},
    }

    assert set(open_computed_rxn.elements) == expected[open_elem.name]


def test_copy(open_computed_rxn):
    assert open_computed_rxn.copy() == open_computed_rxn


def test_reverse(open_computed_rxn):
    open_computed_rxn_rev = open_computed_rxn.reverse()

    assert open_computed_rxn_rev.energy == -open_computed_rxn.energy
    assert open_computed_rxn_rev.reverse() == open_computed_rxn


def test_balance(open_computed_rxn):
    reactant_entries = open_computed_rxn.reactant_entries
    product_entries = open_computed_rxn.product_entries
    chempots = open_computed_rxn.chempots

    rxn = OpenComputedReaction.balance(reactant_entries, product_entries, chempots)

    assert open_computed_rxn == rxn


def test_total_chemical_system(open_computed_rxn):
    open_elem = list(open_computed_rxn.chempots.keys())[0].name

    elems = [e.name for e in open_computed_rxn.elements]
    elems.append(open_elem)

    total_chemsys = "-".join(sorted(set(elems)))
    assert open_computed_rxn.total_chemical_system == total_chemsys
